import sqlite3

from alembic import context
from alembic.operations import ops
from sqlalchemy import create_engine

from benji.config import Config as BenjiConfig
from benji.database import Base

# Only load configuration when we're running standalone
config = context.config
if config.attributes.get('connection', None) is None:
    benji_config = BenjiConfig()
    config.attributes['benji_config'] = benji_config
    database_engine = benji_config.get('databaseEngine', types=str)
    connectable = create_engine(database_engine)
else:
    connectable = config.attributes['connection']

target_metadata = Base.metadata


def run_migrations_online():
    # Sources: https://alembic.sqlalchemy.org/en/latest/cookbook.html#don-t-emit-drop-index-when-the-table-is-to-be-dropped-as-well,
    #          https://alembic.sqlalchemy.org/en/latest/cookbook.html#don-t-generate-empty-migrations-with-autogenerate
    def process_revision_directives(context, revision, directives):
        if config.cmd_opts.autogenerate:
            script = directives[0]
            if not script.upgrade_ops.is_empty():
                # process both "def upgrade()", "def downgrade()"
                for directive in (script.upgrade_ops, script.downgrade_ops):

                    # make a set of tables that are being dropped within
                    # the migration function
                    tables_dropped = set()
                    for op in directive.ops:
                        if isinstance(op, ops.DropTableOp):
                            tables_dropped.add((op.table_name, op.schema))

                    # now rewrite the list of "ops" such that DropIndexOp
                    # is removed for those tables.   Needs a recursive function.
                    directive.ops = list(_filter_drop_indexes(directive.ops, tables_dropped))
            else:
                directives[:] = []

    def _filter_drop_indexes(directives, tables_dropped):
        # given a set of (tablename, schemaname) to be dropped, filter
        # out DropIndexOp from the list of directives and yield the result.

        for directive in directives:
            # ModifyTableOps is a container of ALTER TABLE types of
            # commands.  process those in place recursively.
            if isinstance(directive, ops.ModifyTableOps) and \
                    (directive.table_name, directive.schema) in tables_dropped:
                directive.ops = list(_filter_drop_indexes(directive.ops, tables_dropped))

                # if we emptied out the directives, then skip the
                # container altogether.
                if not directive.ops:
                    continue
            elif isinstance(directive, ops.DropIndexOp) and \
                    (directive.table_name, directive.schema) in tables_dropped:
                # we found a target DropIndexOp.   keep looping
                continue

            # otherwise if not filtered, yield out the directive
            yield directive

    with connectable.connect() as connection:
        if isinstance(connection.connection, sqlite3.Connection):
            connection.execute('PRAGMA foreign_keys=OFF')

        context.configure(connection=connection,
                          target_metadata=target_metadata,
                          compare_type=True,
                          compare_server_default=True,
                          process_revision_directives=process_revision_directives,
                          render_as_batch=True)

        with context.begin_transaction():
            context.run_migrations()

        if isinstance(connection.connection, sqlite3.Connection):
            connection.execute('PRAGMA foreign_keys=ON')


if not context.is_offline_mode():
    run_migrations_online()
else:
    raise RuntimeError('Offline migrations are not supported.')



#!/usr/bin/env python3
import sys

import benji.helpers.prometheus as prometheus
import benji.helpers.settings as settings
import benji.helpers.utils as utils

utils.setup_logging()


def main():
    incomplete_versions = utils.subprocess_run([
        'benji',
        '--machine-output',
        '--log-level',
        settings.benji_log_level,
        'ls',
        'status == "incomplete" and date < "1 day ago"',
    ],
                                               decode_json=True)

    invalid_versions = utils.subprocess_run([
        'benji',
        '--machine-output',
        '--log-level',
        settings.benji_log_level,
        'ls',
        'status == "invalid"',
    ],
                                            decode_json=True)

    prometheus.older_incomplete_versions.set(len(incomplete_versions['versions']))
    prometheus.invalid_versions.set(len(invalid_versions['versions']))
    prometheus.push(prometheus.version_status_registry)
    sys.exit(0)



from collections import namedtuple

import semantic_version

_VersionsTuple = namedtuple('_VersionsTuple', ['configuration', 'database_metadata', 'object_metadata'])
_VersionSpecPair = namedtuple('_VersionTuple', ['current', 'supported'])

# Configuration versions only use the major version part
VERSIONS = _VersionsTuple(configuration=_VersionSpecPair(current=semantic_version.Version('1.0.0'),
                                                         supported=semantic_version.SimpleSpec('>=1,<2')),
                          database_metadata=_VersionSpecPair(current=semantic_version.Version('3.0.0'),
                                                             supported=semantic_version.SimpleSpec('>=1,<4')),
                          object_metadata=_VersionSpecPair(current=semantic_version.Version('2.0.0'),
                                                           supported=semantic_version.SimpleSpec('>=1,<3')))



#!/usr/bin/env python3
import sys
import time
from typing import Optional

import benji.helpers.prometheus as prometheus
import benji.helpers.settings as settings
import benji.helpers.utils as utils

utils.setup_logging()


def main():
    command = ' '.join(sys.argv[1:])
    start_time = time.time()

    prometheus.command_start_time.labels(command=command).set(start_time)
    try:
        utils.subprocess_run(['benji', '--log-level', settings.benji_log_level] + sys.argv[1:])
    except Exception as exception:
        prometheus.command_status_failed.labels(command=command).set(1)
        completion_time = time.time()
        prometheus.command_completion_time.labels(command=command).set(completion_time)
        prometheus.command_runtime_seconds.labels(command=command).set(completion_time - start_time)
        prometheus.push(prometheus.command_registry)
        raise exception
    else:
        prometheus.command_status_succeeded.labels(command=command).set(1)
        completion_time = time.time()
        prometheus.command_completion_time.labels(command=command).set(completion_time)
        prometheus.command_runtime_seconds.labels(command=command).set(completion_time - start_time)
        prometheus.push(prometheus.command_registry)
        sys.exit(0)



import string
from datetime import datetime
from io import StringIO

import colorama


class FormatRenderer:

    def __init__(self, fmt: str, colors: bool = True, force_colors: bool = False):
        if colors is True:
            if force_colors:
                colorama.deinit()
                colorama.init(strip=False)
            else:
                colorama.init()

            self._level_to_color = {
                "critical": colorama.Fore.RED,
                "exception": colorama.Fore.RED,
                "error": colorama.Fore.RED,
                "warn": colorama.Fore.YELLOW,
                "warning": colorama.Fore.YELLOW,
                "info": colorama.Fore.GREEN,
                "debug": colorama.Fore.WHITE,
                "notset": colorama.Back.RED,
            }

            self._reset = colorama.Style.RESET_ALL
        else:
            self._level_to_color = {
                "critical": '',
                "exception": '',
                "error": '',
                "warn": '',
                "warning": '',
                "info": '',
                "debug": '',
                "notset": '',
            }

            self._reset = ''

        self._vformat = string.Formatter().vformat
        self._fmt = fmt

    def __call__(self, _, __, event_dict):
        message = StringIO()

        event_dict['log_color_reset'] = self._reset

        if 'level' in event_dict:
            level = event_dict['level']
            if level in self._level_to_color:
                event_dict['log_color'] = self._level_to_color[level]
            else:
                event_dict['log_color'] = ''
            event_dict['level_uc'] = level.upper()
        else:
            event_dict['log_color'] = ''

        if 'timestamp' in event_dict:
            event_dict['timestamp_local_ctime'] = datetime.fromtimestamp(event_dict['timestamp']).ctime()

        message.write(self._vformat(self._fmt, [], event_dict))

        stack = event_dict.pop("stack", None)
        exception = event_dict.pop("exception", None)

        if stack is not None:
            message.write("\n" + stack)
        if exception is not None:
            message.write("\n" + exception)

        message.write(self._reset)

        return message.getvalue()



"""Add table storages

Revision ID: dd844d630d49
Revises: 2bb97229fe36
Create Date: 2019-10-01 00:06:35.657495

"""
import sqlalchemy as sa
from alembic import op

from benji.config import Config

# revision identifiers, used by Alembic.
revision = 'dd844d630d49'
down_revision = '2bb97229fe36'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table('storages', sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
                    sa.Column('name', sa.String(length=255), nullable=False),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_storages')),
                    sa.UniqueConstraint('name', name=op.f('uq_storages_name')))

    benji_config = op.get_context().config.attributes.get('benji_config', None)
    if benji_config is not None:
        storages = sa.Table('storages', sa.MetaData(), autoload_with=op.get_bind())
        storages_list = benji_config.get('storages', types=list)
        for index, storage in enumerate(storages_list):
            name = Config.get_from_dict(storage, 'name', types=str, index=index)
            storage_id = Config.get_from_dict(storage, 'storageId', None, types=int, index=index)
            op.execute(storages.insert().values(name=name, id=storage_id))

    with op.batch_alter_table('versions', schema=None) as batch_op:
        batch_op.create_foreign_key(batch_op.f('fk_versions_storage_id_storages'), 'storages', ['storage_id'], ['id'])
    with op.batch_alter_table('deleted_blocks', schema=None) as batch_op:
        batch_op.create_foreign_key(batch_op.f('fk_deleted_blocks_storage_id_storages'), 'storages', ['storage_id'],
                                    ['id'])


def downgrade():
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import threading
import time
from typing import Tuple, Optional, Callable, Any, List, Union, Iterator

# noinspection PyUnresolvedReferences
import libiscsi.libiscsi as libiscsi

from benji.config import ConfigDict, Config
from benji.database import DereferencedBlock, Block
from benji.exception import ConfigurationError, UsageError
from benji.io.base import IOBase
from benji.logging import logger


class IO(IOBase):

    _pool_name: Optional[str]
    _image_name: Optional[str]
    _snapshot_name: Optional[str]

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict, url: str,
                 block_size: int) -> None:
        super().__init__(config=config,
                         name=name,
                         module_configuration=module_configuration,
                         url=url,
                         block_size=block_size)

        if self.parsed_url.params or self.parsed_url.fragment:
            raise UsageError('The supplied URL {} is invalid.'.format(self.url))

        self._read_queue: List[DereferencedBlock] = []
        self._outstanding_write: Optional[Tuple[DereferencedBlock, bytes]] = None

        self._username = config.get_from_dict(module_configuration, 'username', None, types=str)
        self._password = config.get_from_dict(module_configuration, 'password', None, types=str)
        self._target_username = config.get_from_dict(module_configuration, 'targetUsername', None, types=str)
        self._target_password = config.get_from_dict(module_configuration, 'targetPassword', None, types=str)
        header_digest = config.get_from_dict(module_configuration, 'headerDigest', types=str)
        header_digest_attr_name = 'ISCSI_HEADER_DIGEST_{}'.format(header_digest)
        if hasattr(libiscsi, header_digest_attr_name):
            self._header_digest = getattr(libiscsi, header_digest_attr_name)
        else:
            raise ConfigurationError('Unknown header digest setting {}.'.format(header_digest))
        self._initiator_name = config.get_from_dict(module_configuration, 'initiatorName', types=str)
        self._timeout = config.get_from_dict(module_configuration, 'timeout', None, types=int)

        self._iscsi_context: Any = None

    @staticmethod
    def _iscsi_execute_sync(operation: str, function: Callable, iscsi_context, *args, **kwargs) -> Any:
        task = function(iscsi_context, *args, **kwargs)
        if task is not None:
            status, sense = libiscsi.scsi_task_get_status(task)
            if status != libiscsi.SCSI_STATUS_GOOD:
                raise RuntimeError('{} failed with {}, ASCQ {}.'.format(operation,
                                                                        libiscsi.scsi_sense_key_str(sense.key),
                                                                        libiscsi.scsi_sense_ascq_str(sense.ascq)))
            else:
                return task
        else:
            raise RuntimeError('{} failed: {}'.format(operation, libiscsi.iscsi_get_error(iscsi_context)))

    @staticmethod
    def _iscsi_call_sync(operation: str, function: Callable, iscsi_context, *args, **kwargs) -> Any:
        result = function(iscsi_context, *args, **kwargs)
        if result is None or isinstance(result, int) and result < 0:
            raise RuntimeError('{} failed: {}'.format(operation, libiscsi.iscsi_get_error(iscsi_context).rstrip()))
        return result

    def _open(self) -> None:
        if self._iscsi_context is not None:
            return

        # netloc includes username, password and port
        url = 'iscsi://{}{}?{}'.format(self.parsed_url.netloc, self.parsed_url.path, self.parsed_url.query)
        iscsi_context = libiscsi.iscsi_create_context(self._initiator_name)
        iscsi_url = self._iscsi_call_sync('URL parsing', libiscsi.iscsi_parse_full_url, self._iscsi_context, url)

        libiscsi.iscsi_set_targetname(iscsi_context, iscsi_url.target)
        libiscsi.iscsi_set_session_type(iscsi_context, libiscsi.ISCSI_SESSION_NORMAL)
        libiscsi.iscsi_set_header_digest(iscsi_context, self._header_digest)
        libiscsi.iscsi_set_timeout(iscsi_context, self._timeout)

        if len(iscsi_url.user) > 0:
            libiscsi.iscsi_set_initiator_username_pwd(iscsi_context, iscsi_url.user, iscsi_url.passwd)
        elif self._username is not None:
            libiscsi.iscsi_set_initiator_username_pwd(iscsi_context, self._username, self._password)
        # iscsi_set_target_username_pwd needs to come after iscsi_set_initiator_username_pwd
        if len(iscsi_url.target_user) > 0:
            libiscsi.iscsi_set_target_username_pwd(iscsi_context, iscsi_url.target_user, iscsi_url.target_passwd)
        elif self._target_username is not None:
            libiscsi.iscsi_set_target_username_pwd(iscsi_context, self._target_username, self._target_password)

        self._iscsi_call_sync('iSCSI connect', libiscsi.iscsi_full_connect_sync, iscsi_context, iscsi_url.portal,
                              iscsi_url.lun)

        task = self._iscsi_execute_sync('READ CAPACITY(16)', libiscsi.iscsi_readcapacity16_sync, iscsi_context,
                                        iscsi_url.lun)
        task_data = libiscsi.scsi_datain_unmarshall(task)
        if not task_data:
            raise RuntimeError('Failed to unmarshall READ CAPACITY(16) data.')

        self._iscsi_block_size = task_data.readcapacity16.block_length
        self._iscsi_num_blocks = task_data.readcapacity16.returned_lba + 1
        self._fully_provisioned = task_data.readcapacity16.lbpme == 0
        self._unmapped_is_zero = task_data.readcapacity16.lbprz != 0
        self._iscsi_lun = iscsi_url.lun

        if self.block_size < self._iscsi_block_size:
            raise RuntimeError('Block size of version is smaller than block size of iSCSI target ({} < {}).'.format(
                self.block_size, self._iscsi_block_size))

        if self.block_size % self._iscsi_block_size != 0:
            raise RuntimeError('Block size of version is not aligned to block size of iSCSI target (remainder {}).'.format(
                self.block_size % self._iscsi_block_size))

        self._iscsi_context = iscsi_context

        logger.debug('Opened iSCSI device {} with block size {} and {} blocks.'.format(
            self.url, self._iscsi_block_size, self._iscsi_num_blocks))

    def open_r(self) -> None:
        self._open()

    def open_w(self, size: int, force: bool = False, sparse: bool = False) -> None:
        self._open()

    def close(self) -> None:
        if len(self._read_queue) > 0:
            logger.warning('Closing IO module with {} outstanding read jobs.'.format(len(self._read_queue)))
            self._read_queue = []

        if self._outstanding_write is not None:
            logger.warning('Closing IO module with one outstanding write.')
            self._outstanding_write = None

        self._iscsi_context = None

    def size(self) -> int:
        self._open()
        return self._iscsi_block_size * self._iscsi_num_blocks

    def _read(self, block: DereferencedBlock) -> Tuple[DereferencedBlock, bytes]:
        assert block.size == self.block_size
        lba = (block.idx * self.block_size) // self._iscsi_block_size
        num_blocks = self.block_size // self._iscsi_block_size

        if lba >= self._iscsi_num_blocks:
            raise RuntimeError(
                'Attempt to read outside of the device. Requested LBA is {}, but device has only {} blocks. (1)'.format(
                    lba, self._iscsi_num_blocks))

        if lba + num_blocks > self._iscsi_num_blocks:
            raise RuntimeError(
                'Attempt to read outside of the device. Requested LBA is {}, but device has only {} blocks. (2)'.format(
                    lba + num_blocks, self._iscsi_num_blocks))

        t1 = time.time()
        task = self._iscsi_execute_sync('READ(16)', libiscsi.iscsi_read16_sync, self._iscsi_context, self._iscsi_lun,
                                        lba, self.block_size, self._iscsi_block_size, 0, 0, 0, 0, 0)

        data = task.datain
        assert len(data) == self.block_size
        t2 = time.time()

        logger.debug('{} read block {} in {:.3f}s'.format(
            threading.current_thread().name,
            block.idx,
            t2 - t1,
        ))

        return block, data

    def read(self, block: Union[DereferencedBlock, Block]) -> None:
        self._read_queue.append(block.deref())

    def read_sync(self, block: Union[DereferencedBlock, Block]) -> bytes:
        return self._read(block.deref())[1]

    def read_get_completed(self, timeout: Optional[int] = None
                          ) -> Iterator[Union[Tuple[DereferencedBlock, bytes], BaseException]]:
        while len(self._read_queue) > 0:
            yield self._read(self._read_queue.pop())

    def _write(self, block: DereferencedBlock, data: bytes) -> DereferencedBlock:
        assert block.size == self.block_size
        lba = (block.idx * self.block_size) // self._iscsi_block_size
        num_blocks = self.block_size // self._iscsi_block_size

        if lba >= self._iscsi_num_blocks:
            raise RuntimeError(
                'Attempt to write outside of the device. Requested LBA is {}, but device has only {} blocks. (1)'.format(
                    lba, self._iscsi_num_blocks))

        if lba + num_blocks > self._iscsi_num_blocks:
            raise RuntimeError(
                'Attempt to write outside of the device. Requested LBA is {}, but device has only {} blocks. (2)'.format(
                    lba + num_blocks, self._iscsi_num_blocks))

        t1 = time.time()
        self._iscsi_execute_sync('WRITE(16)', libiscsi.iscsi_write16_sync, self._iscsi_context, self._iscsi_lun, lba,
                                 data, self._iscsi_block_size, 0, 0, 0, 0, 0)
        t2 = time.time()

        logger.debug('{} wrote block {} in {:.3f}s'.format(
            threading.current_thread().name,
            block.idx,
            t2 - t1,
        ))

        return block

    def write(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        assert self._outstanding_write is None
        self._outstanding_write = (block.deref(), data)

    def write_sync(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        self._write(block.deref(), data)

    def write_get_completed(self, timeout: Optional[int] = None) -> Iterator[Union[DereferencedBlock, BaseException]]:
        if self._outstanding_write is not None:
            yield self._write(*self._outstanding_write)
            self._outstanding_write = None



import functools
import json
from io import StringIO
from typing import List, Optional, Dict, Any

from bottle import Bottle, response, HTTPError, request
from webargs import fields
from webargs.bottleparser import use_kwargs

import benji.exception
from benji import __version__
from benji.benji import Benji
from benji.config import Config
from benji.database import Version, VersionUid
from benji.utils import hints_from_rbd_diff, InputValidation, random_string
from benji.versions import VERSIONS


def error(status: int):

    def decorator(func):
        func.bottle_error = status

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            body = json.dumps(func(*args, **kwargs))
            response.content_type = 'application/json; charset=utf-8'
            return body

        return wrapped_func()

    return decorator


def route(path: str, **decorator_kwargs):

    def decorator(func):

        func.bottle_route = decorator_kwargs
        func.bottle_route['path'] = path

        annotations = getattr(func, "__annotations__", {})
        request_args = {
            name: value for name, value in annotations.items() if isinstance(value, fields.Field) and name != "return"
        }
        func.bottle_route['apply'] = use_kwargs(request_args)

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            body = func(*args, **kwargs)

            if response.status_code == 200 and body is None:
                response.status = 204

            if isinstance(body, StringIO):
                response.content_type = 'application/json; charset=utf-8'
                return body.getvalue()
            else:
                response.content_type = 'application/json; charset=utf-8'
                return json.dumps(body) if body is not None else ''

        return wrapped_func

    return decorator


class RestAPI:
    """Proxy between REST calls and actual backup code."""

    CORE_API_VERSION_V1 = 'v1'
    CORE_API_GROUP = 'core'

    def __init__(self, config: Config):
        self._app = Bottle()

        def default_error_handler(error: HTTPError):
            return self._default_error_handler(error)

        self._app.default_error_handler = default_error_handler
        self._config = config
        self._install_routes()

    def _install_routes(self):
        for kw in dir(self):
            attr = getattr(self, kw)
            if hasattr(attr, 'bottle_route'):
                self._app.route(**attr.bottle_route)(attr)
            if hasattr(attr, 'bottle_error'):
                self._app.error(attr.bottle_error)(attr)

    def run(self, bind_address: str, bind_port: int, threads: int, debug: bool = False):
        self._app.run(
            server='gunicorn',
            host=bind_address,
            port=bind_port,
            reuse_port=True,
            worker_class='gthread',
            threads=threads,
            worker_connections=threads * 10,
            debug=debug,
        )

    @staticmethod
    def _default_error_handler(error: HTTPError):

        result = {
            'url': repr(request.url),
            'status': error.status,
            'body': error.body,
        }

        if error.exception:
            result['exception'] = {'name': error.exception.__class__.__name__, 'message': str(error.exception)}

        if error.traceback:
            result['traceback'] = error.traceback

        response.content_type = 'application/json; charset=utf-8'
        return json.dumps(result)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions', method='POST')
    def _api_v1_versions_create(
        self, version_uid: fields.Str(missing=None), volume: fields.Str(required=True),
        snapshot: fields.Str(required=True), source: fields.Str(required=True), rbd_hints: fields.Str(missing=None),
        base_version_uid: fields.Str(missing=None), block_size: fields.Int(missing=None),
        storage_name: fields.Str(missing=None)
    ) -> StringIO:
        if version_uid is None:
            version_uid = '{}-{}'.format(volume[:248], random_string(6))
        version_uid_obj = VersionUid(version_uid)
        base_version_uid_obj = VersionUid(base_version_uid) if base_version_uid else None

        result = StringIO()
        with Benji(self._config) as benji_obj:
            hints = None
            if rbd_hints:
                with open(rbd_hints, 'r') as f:
                    hints = hints_from_rbd_diff(f.read())
            backup_version = benji_obj.backup(version_uid=version_uid_obj,
                                              volume=volume,
                                              snapshot=snapshot,
                                              source=source,
                                              hints=hints,
                                              base_version_uid=base_version_uid_obj,
                                              storage_name=storage_name,
                                              block_size=block_size)

            benji_obj.export_any({'versions': [backup_version]},
                                 result,
                                 ignore_relationships=[((Version,), ('blocks',))])

            response.status = 201
            response.set_header('Location', f'{request.path}/{backup_version.uid}')
        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>/restore', method='POST')
    def _api_v1_versions_restore_create(
        self, version_uid: str, destination: fields.Str(required=True), sparse: fields.Bool(missing=False),
        force: fields.Bool(missing=False), database_backend_less: fields.Bool(missing=False)
    ) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        result = StringIO()
        with Benji(self._config, in_memory_database=database_backend_less) as benji_obj:
            try:
                if database_backend_less:
                    benji_obj.metadata_restore([version_uid_obj])
                benji_obj.restore(version_uid_obj, destination, sparse, force)

                benji_obj.export_any({'versions': [version_uid_obj]},
                                     result,
                                     ignore_relationships=[((Version,), ('blocks',))])
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>', method='GET')
    def _api_v1_versions_read(self, version_uid: str) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        result = StringIO()
        with Benji(self._config) as benji_obj:
            try:
                benji_obj.export_any({'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]},
                                     result,
                                     ignore_relationships=[((Version,), ('blocks',))])
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>', method='PATCH')
    def _api_v1_versions_patch(
        self, version_uid: str, protected: fields.Bool(missing=None), labels: fields.DelimitedList(fields.Str(),
                                                                                                   missing=None)
    ) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        if labels is not None:
            label_add, label_remove = InputValidation.parse_and_validate_labels(labels)
        else:
            label_add, label_remove = [], []
        result = StringIO()
        with Benji(self._config) as benji_obj:
            try:
                if protected is not None:
                    benji_obj.protect(version_uid_obj, protected=protected)

                for name, value in label_add:
                    benji_obj.add_label(version_uid_obj, name, value)
                for name in label_remove:
                    benji_obj.rm_label(version_uid_obj, name)

                benji_obj.export_any({'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]},
                                     result,
                                     ignore_relationships=[((Version,), ('blocks',))])
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>', method='DELETE')
    def _api_v1_versions_delete(
        self, version_uid: str, force: fields.Bool(missing=False), keep_metadata_backup: fields.Bool(missing=False),
        override_lock: fields.Bool(missing=False)
    ) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        result = StringIO()
        with Benji(self._config) as benji_obj:
            try:
                # Do this before deleting the version
                benji_obj.export_any({'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]},
                                     result,
                                     ignore_relationships=[((Version,), ('blocks',))])

                benji_obj.rm(version_uid_obj,
                             force=force,
                             keep_metadata_backup=keep_metadata_backup,
                             override_lock=override_lock)
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>/scrub', method='POST')
    def _api_v1_versions_scrub_create(self, version_uid: str, block_percentage: fields.Int(missing=100)) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        result = StringIO()
        with Benji(self._config) as benji_obj:
            try:
                try:
                    benji_obj.scrub(version_uid_obj, block_percentage=block_percentage)
                except benji.exception.ScrubbingError:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]
                        },
                        result,
                        ignore_relationships=[((Version,), ('blocks',))])
                else:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': []
                        },
                        result,
                        ignore_relationships=[((Version,), ('blocks',))])
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/<version_uid>/deep-scrub', method='POST')
    def _api_v1_versions_deep_scrub_create(
            self, version_uid: str, source: fields.Str(missing=None),
            block_percentage: fields.Int(missing=100)) -> StringIO:
        version_uid_obj = VersionUid(version_uid)
        result = StringIO()
        with Benji(self._config) as benji_obj:
            try:
                try:
                    benji_obj = Benji(self._config)
                    benji_obj.deep_scrub(version_uid_obj, source=source, block_percentage=block_percentage)
                except benji.exception.ScrubbingError:
                    assert benji_obj is not None
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)]
                        },
                        result,
                        ignore_relationships=[((Version,), ('blocks',))])
                else:
                    benji_obj.export_any(
                        {
                            'versions': [benji_obj.get_version_by_uid(version_uid=version_uid_obj)],
                            'errors': []
                        },
                        result,
                        ignore_relationships=[((Version,), ('blocks',))])
            except KeyError:
                response.status = f'410 Version {version_uid} not found.'

        return result

    def _batch_scrub(self, method: str, filter_expression: Optional[str], version_percentage: int,
                     block_percentage: int, group_label: Optional[str]) -> StringIO:
        with Benji(self._config) as benji_obj:
            versions, errors = getattr(benji_obj, method)(filter_expression, version_percentage, block_percentage,
                                                          group_label)

            result = StringIO()
            benji_obj.export_any({
                'versions': versions,
                'errors': errors,
            },
                                 result,
                                 ignore_relationships=[((Version,), ('blocks',))])

            return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/scrub', method='POST')
    def _api_v1_versions_batch_scrub_create(
        self, filter_expression: fields.Str(missing=None), version_percentage: fields.Int(missing=100),
        block_percentage: fields.Int(missing=100), group_label: fields.Str(missing=None)
    ) -> StringIO:
        return self._batch_scrub('batch_scrub', filter_expression, version_percentage, block_percentage, group_label)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/deep-scrub', method='POST')
    def _api_v1_versions_batch_deep_scrub_create(
        self, filter_expression: fields.Str(missing=None), version_percentage: fields.Int(missing=100),
        block_percentage: fields.Int(missing=100), group_label: fields.Str(missing=None)
    ) -> StringIO:
        return self._batch_scrub('batch_deep_scrub', filter_expression, version_percentage, block_percentage,
                                 group_label)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions', method='GET')
    def _api_v1_versions_list(
            self, filter_expression: fields.Str(missing=None), include_blocks: fields.Bool(missing=False)) -> StringIO:
        with Benji(self._config) as benji_obj:
            versions = benji_obj.find_versions_with_filter(filter_expression)

            result = StringIO()
            benji_obj.export_any(
                {'versions': versions},
                result,
                ignore_relationships=[((Version,), ('blocks',) if not include_blocks else ())],
            )

            return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/blocks', method='DELETE')
    def _api_v1_blocks_delete_collection(self, override_lock: fields.Bool(missing=False)) -> None:
        with Benji(self._config) as benji_obj:
            benji_obj.cleanup(override_lock=override_lock)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/metadata/backup', method='POST')
    def _api_v1_versions_metadata_backup_create(
            self, filter_expression: fields.Str(missing=None), force: fields.Bool(missing=False)) -> None:
        with Benji(self._config) as benji_obj:
            version_uid_objs = [version.uid for version in benji_obj.find_versions_with_filter(filter_expression)]
            benji_obj.metadata_backup(version_uid_objs, overwrite=force)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/metadata/import', method='POST')
    def _api_v1_versions_metadata_import_create(self) -> None:
        with Benji(self._config) as benji_obj:
            benji_obj.metadata_import(request.body.read())

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions/metadata/restore', method='POST')
    def _api_v1_versions_metadata_restore_create(
        self, version_uids: fields.DelimitedList(fields.Str, required=True), storage_name: fields.Str(missing=None)
    ) -> None:
        version_uid_objs = [VersionUid(version_uid) for version_uid in version_uids]
        with Benji(self._config) as benji_obj:
            benji_obj.metadata_restore(version_uid_objs, storage_name)

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/storages', method='GET')
    def _api_v1_storages_list(self) -> List[str]:
        with Benji(self._config) as benji_obj:
            return benji_obj.list_storages()

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/database', method='POST')
    def _api_v1_database_init_create(self) -> None:
        Benji(self._config, init_database=True).close()

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/database', method='PATCH')
    def _api_v1_database_migrate_create(self) -> None:
        Benji(self._config, migrate_database=True).close()

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/versions', method='DELETE')
    def _api_v1_versions_delete_collection(
        self, rules_spec: fields.Str(required=True), filter_expression: fields.Str(missing=None),
        dry_run: fields.Bool(missing=False), keep_metadata_backup: fields.Bool(missing=False),
        group_label: fields.Str(missing=None)
    ) -> StringIO:
        with Benji(self._config) as benji_obj:
            dismissed_versions = benji_obj.enforce_retention_policy(filter_expression=filter_expression,
                                                                    rules_spec=rules_spec,
                                                                    dry_run=dry_run,
                                                                    keep_metadata_backup=keep_metadata_backup,
                                                                    group_label=group_label)

            result = StringIO()
            benji_obj.export_any(
                {'versions': dismissed_versions},
                result,
                ignore_relationships=[((Version,), ('blocks',))],
            )

            return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/version-info', method='GET')
    def _api_v1_version_info_read(self) -> Dict[str, Any]:
        result = {
            'version': __version__,
            'configuration_version': {
                'current': str(VERSIONS.configuration.current),
                'supported': str(VERSIONS.configuration.supported)
            },
            'database_metadata_version': {
                'current': str(VERSIONS.database_metadata.current),
                'supported': str(VERSIONS.database_metadata.supported)
            },
            'object_metadata_version': {
                'current': str(VERSIONS.object_metadata.current),
                'supported': str(VERSIONS.object_metadata.supported)
            },
        }

        return result

    @route(f'/apis/{CORE_API_GROUP}/{CORE_API_VERSION_V1}/storages/<storage_name>', method='GET')
    def _api_v1_storages_read(self, storage_name: str) -> Dict[str, int]:
        with Benji(self._config) as benji_obj:
            objects_count, objects_size = benji_obj.storage_stats(storage_name)

            result = {
                'objects_count': objects_count,
                'objects_size': objects_size,
            }

            return result



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from abc import ABCMeta, abstractmethod
from typing import Tuple, Union, Optional, Iterator
from urllib import parse

from benji.config import ConfigDict, Config
from benji.database import Block, DereferencedBlock
from benji.repr import ReprMixIn


class IOBase(ReprMixIn, metaclass=ABCMeta):

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict, url: str,
                 block_size: int) -> None:
        self._name = name
        self._url = url
        self._parsed_url = parse.urlparse(url, allow_fragments=False)
        self._block_size = block_size

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> str:
        return self._url

    @property
    def parsed_url(self) -> parse.ParseResult:
        return self._parsed_url

    @property
    def sanitized_url(self) -> str:
        # This assumes that sensitive information is either contained in the username, password, params,
        # query or fragment part of the URL. If not, this property could be overridden by the subclass.
        return parse.ParseResult(scheme=self._parsed_url.scheme,
                                 netloc=self._parsed_url.netloc.split('@')[-1],
                                 path=self._parsed_url.path,
                                 params='',
                                 query='',
                                 fragment='').geturl()

    @property
    def block_size(self):
        return self._block_size

    @abstractmethod
    def open_r(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def open_w(self, size: int, force: bool = False, sparse: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def read(self, block: Union[DereferencedBlock, Block]) -> None:
        raise NotImplementedError

    @abstractmethod
    def read_sync(self, block: Union[DereferencedBlock, Block]) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def read_get_completed(self,
                           timeout: Optional[int] = None
                          ) -> Iterator[Union[Tuple[DereferencedBlock, bytes], BaseException]]:
        raise NotImplementedError

    @abstractmethod
    def write(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        raise NotImplementedError

    @abstractmethod
    def write_sync(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        raise NotImplementedError

    @abstractmethod
    def write_get_completed(self, timeout: Optional[int] = None) -> Iterator[Union[DereferencedBlock, BaseException]]:
        raise NotImplementedError



"""volume_version_uid

Revision ID: da25cc147a07
Revises: dd844d630d49
Create Date: 2019-10-28 15:20:15.455215

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'da25cc147a07'
down_revision = 'dd844d630d49'
branch_labels = None
depends_on = None


def upgrade():
    # Constraint names need to be globally unique with PostgreSQL (and they share the same namespace as tables)
    # For the benefit of PostgreSQL drop the primary key constraints before recreating them below to avoid collisions.
    # Requires PostgreSQL 9.2 or newer.
    if (op.get_context().dialect.name == 'postgresql'):
        op.execute('ALTER TABLE versions RENAME CONSTRAINT "pk_versions" TO "pk_versions_old"')
        op.execute('ALTER TABLE blocks RENAME CONSTRAINT "pk_blocks" TO "pk_blocks_old"')
        op.execute('ALTER TABLE labels RENAME CONSTRAINT "pk_labels" TO "pk_labels_old"')

    op.create_table('versions_new',
                    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
                    sa.Column('uid', sa.String(length=255), nullable=False),
                    sa.Column('date', sa.DateTime(), nullable=False),
                    sa.Column('volume', sa.String(length=255), nullable=False),
                    sa.Column('snapshot', sa.String(length=255), nullable=False),
                    sa.Column('size', sa.BigInteger(), nullable=False),
                    sa.Column('block_size', sa.Integer(), nullable=False),
                    sa.Column('storage_id', sa.Integer(), nullable=False),
                    sa.Column('status', sa.Integer(), nullable=False),
                    sa.Column('protected', sa.Boolean(name=op.f('ck_versions_protected')), nullable=False),
                    sa.Column('bytes_read', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_written', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_dedup', sa.BigInteger(), nullable=True),
                    sa.Column('bytes_sparse', sa.BigInteger(), nullable=True),
                    sa.Column('duration', sa.BigInteger(), nullable=True),
                    sa.ForeignKeyConstraint(['storage_id'], ['storages.id'],
                                            name=op.f('fk_versions_storage_id_storages')),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_versions')),
                    sa.UniqueConstraint('uid', name=op.f('uq_versions_uid')),
                    sqlite_autoincrement=True)
    with op.batch_alter_table('versions_new', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_versions_volume'), ['volume'], unique=False)

    conn = op.get_bind()
    metadata = sa.MetaData()
    versions = sa.Table('versions', metadata, autoload_with=conn)
    versions_new = sa.Table('versions_new', metadata, autoload_with=conn)

    for version in conn.execute(versions.select()):
        conn.execute(versions_new.insert().values(id=version.uid,
                                                  uid=f'V{version.uid:010d}',
                                                  date=version.date,
                                                  volume=version.name,
                                                  snapshot=version.snapshot,
                                                  size=version.size,
                                                  block_size=version.block_size,
                                                  storage_id=version.storage_id,
                                                  status=version.status,
                                                  protected=version.protected,
                                                  bytes_read=version.bytes_read,
                                                  bytes_written=version.bytes_written,
                                                  bytes_dedup=version.bytes_dedup,
                                                  bytes_sparse=version.bytes_sparse,
                                                  duration=version.duration))

    # Reset sequence for PostgreSQL
    # Source: https://stackoverflow.com/questions/244243/how-to-reset-postgres-primary-key-sequence-when-it-falls-out-of-sync
    if (op.get_context().dialect.name == 'postgresql'):
        op.execute('SELECT setval(pg_get_serial_sequence(\'versions_new\', \'id\'), COALESCE(MAX(id), 0) + 1, false) FROM versions_new')

    with op.batch_alter_table('blocks', schema=None) as batch_op:
        batch_op.drop_constraint('fk_blocks_version_uid_versions', type_='foreignkey')

    with op.batch_alter_table('labels', schema=None) as batch_op:
        batch_op.drop_constraint('fk_labels_version_uid_versions', type_='foreignkey')

    op.drop_table('versions')
    op.rename_table('versions_new', 'versions')
    # Remove old table schema from metadata and reload it (keep_existing=False didn't work for some reason)
    metadata.remove(versions)
    versions = sa.Table('versions', metadata, autoload_with=conn)

    # Blocks

    op.create_table(
        'blocks_new', sa.Column('idx', sa.Integer(), nullable=False), sa.Column('uid_right',
                                                                                sa.Integer(),
                                                                                nullable=True),
        sa.Column('uid_left', sa.Integer(), nullable=True), sa.Column('size', sa.Integer(), nullable=True),
        sa.Column('version_id', sa.Integer(), nullable=False),
        sa.Column('valid', sa.Boolean(name=op.f('ck_blocks_valid')), nullable=False),
        sa.Column('checksum', sa.LargeBinary(length=64), nullable=True),
        sa.ForeignKeyConstraint(['version_id'], ['versions.id'],
                                name=op.f('fk_blocks_version_id_versions'),
                                ondelete='CASCADE'), sa.PrimaryKeyConstraint('version_id',
                                                                             'idx',
                                                                             name=op.f('pk_blocks')))

    with op.batch_alter_table('blocks', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_blocks_checksum'))
        batch_op.drop_index(batch_op.f('ix_blocks_uid_left'))
    with op.batch_alter_table('blocks_new', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_blocks_checksum'), ['checksum'], unique=False)
        batch_op.create_index(batch_op.f('ix_blocks_uid_left'), ['uid_left', 'uid_right'], unique=False)

    blocks = sa.Table('blocks', metadata, autoload_with=conn)
    blocks_new = sa.Table('blocks_new', metadata, autoload_with=conn)

    for block in conn.execute(blocks.select()):
        conn.execute(blocks_new.insert().values(idx=block.idx,
                                                uid_left=block.uid_left,
                                                uid_right=block.uid_right,
                                                size=block.size,
                                                version_id=block.version_uid,
                                                valid=block.valid,
                                                checksum=block.checksum))

    op.drop_table('blocks')
    op.rename_table('blocks_new', 'blocks')

    # Labels

    op.create_table(
        'labels_new', sa.Column('version_id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['version_id'], ['versions.id'],
                                name=op.f('fk_labels_version_id_versions'),
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('version_id', 'name', name=op.f('pk_labels')))
    with op.batch_alter_table('labels', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_labels_value'))
    with op.batch_alter_table('labels_new', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_labels_name'), ['name'], unique=False)
        batch_op.create_index(batch_op.f('ix_labels_value'), ['value'], unique=False)

    labels = sa.Table('labels', metadata, autoload_with=conn)
    labels_new = sa.Table('labels_new', metadata, autoload_with=conn)

    for label in conn.execute(labels.select()):
        conn.execute(labels_new.insert().values(name=label.name, value=label.value, version_id=label.version_uid))

    op.drop_table('labels')
    op.rename_table('labels_new', 'labels')


def downgrade():
    pass



from benji.utils import BlockHash


def test_sha512():
    bh = BlockHash('SHA512')
    assert 'daef4953b9783365cad6615223720506cc46c5167cd16ab500fa597aa08ff964eb24fb19687f34d7665f778fcb6c5358fc0a5b81e1662cf90f73a2671c53f991' == bh.data_hexdigest(
        b'test123')


def test_blake2_16():
    bh = BlockHash('BLAKE2b,digest_bits=128')
    assert '6de7714a67685c8f448db98d3d1a307a' == bh.data_hexdigest(b'test123')


def test_blake2_32():
    bh = BlockHash('BLAKE2b,digest_bits=256')
    assert '90cccd774db0ac8c6ea2deff0e26fc52768a827c91c737a2e050668d8c39c224' == bh.data_hexdigest(b'test123')



import sys
from typing import Any, Dict, Tuple, List, Optional
from urllib.error import HTTPError

import requests

# Use a session so that this client will work seamlessly with cookies.
requests_session = requests.Session()


class BenjiRESTClient:

    CORE_API_VERSION_V1 = 'v1'
    CORE_API_GROUP = 'core'

    def __init__(self, api_endpoint: str):
        self._api_endpoint = api_endpoint

    def _api_request(self,
                     path: str,
                     method: str = 'GET',
                     body: Any = None,
                     params: Dict[str, Any] = None,
                     timeout: Tuple[int, int] = (2, 30),
                     api_version: str = CORE_API_VERSION_V1,
                     api_group=CORE_API_GROUP) -> Optional[Dict[str, Any]]:
        response = requests_session.request(method,
                                            f'{self._api_endpoint}/apis/{api_group}/{api_version}/{path}',
                                            headers={'Content-Type': 'application/json; charset=utf-8'},
                                            params=params,
                                            json=body,
                                            timeout=timeout)

        if response.status_code in (404, 410):
            raise KeyError(response.reason)

        # Raise for anything != 2xx
        response.raise_for_status()

        if response.status_code != 204:
            return response.json()
        else:
            return None

    def get_version_by_uid(self, version_uid: str = None) -> Dict[str, Any]:
        return self._api_request(f'versions/{version_uid}')['versions'][0]

    def find_versions_with_filter(self,
                                  filter_expression: str = None,
                                  include_blocks: bool = False) -> List[Dict[str, Any]]:
        return self._api_request('versions',
                                 params={
                                     'filter_expression': filter_expression,
                                     'include_blocks': include_blocks
                                 })['versions']

    def rm(self,
           version_uid: str,
           force: bool = False,
           keep_metadata_backup: bool = False,
           override_lock: bool = False):
        return self._api_request(f'versions/{version_uid}',
                                 method='DELETE',
                                 params={
                                     'force': force,
                                     'keep_metadata_backup': keep_metadata_backup,
                                     'override_lock': override_lock
                                 })['versions'][0]

    def protect(self, version_uid: str, protected: bool):
        return self._api_request(f'versions/{version_uid}', method='PATCH',
                                 params={'protected': protected})['versions'][0]



# -*- encoding: utf-8 -*-

import datetime
import errno
import hashlib
import os
import random
import time
from collections import defaultdict
from concurrent.futures import CancelledError, TimeoutError
from contextlib import AbstractContextManager
from io import StringIO, BytesIO
from typing import List, Tuple, TextIO, Optional, Set, Dict, cast, Union, \
    Sequence, Any, Iterator

from benji.blockuidhistory import BlockUidHistory
from benji.config import Config
from benji.database import Database, VersionUid, Version, Block, \
    BlockUid, DereferencedBlock, VersionStatus, Storage, Locking, DeletedBlock, SparseBlockUid
from benji.exception import InputDataError, InternalError, AlreadyLocked, UsageError, ScrubbingError, ConfigurationError
from benji.io.factory import IOFactory
from benji.logging import logger
from benji.repr import ReprMixIn
from benji.retentionfilter import RetentionFilter
from benji.storage.base import InvalidBlockException, BlockNotFoundError
from benji.storage.factory import StorageFactory
from benji.utils import notify, BlockHash, PrettyPrint, random_string, InputValidation
from diskcache import Cache


class Benji(ReprMixIn, AbstractContextManager):

    # This is in number of blocks (i.e. database rows in the blocks table)
    _BLOCKS_CREATE_WORK_PACKAGE = 10000

    def __init__(self,
                 config: Config,
                 init_database: bool = False,
                 migrate_database: bool = False,
                 in_memory_database: bool = False,
                 _destroy_database: bool = False) -> None:

        self.config = config

        self._block_size = config.get('blockSize', types=int)
        self._block_hash = BlockHash(config.get('hashFunction', types=str))
        self._process_name = config.get('processName', types=str)

        Database.configure(config, in_memory=in_memory_database)
        if init_database or in_memory_database:
            Database.init(_destroy=_destroy_database)
        if migrate_database:
            Database.migrate()
        Database.open()

        # Ensure that all defined storages are present
        storage_modules = StorageFactory.get_modules()
        for name, module in storage_modules.items():
            Storage.sync(name, storage_id=module.storage_id)
        default_storage_name = self.config.get('defaultStorage', types=str)
        if default_storage_name not in storage_modules.keys():
            raise ConfigurationError('Default storage {} is undefined.'.format(default_storage_name))
        self._default_storage_name = default_storage_name

        notify(self._process_name)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def create_version(self,
                       *,
                       version_uid: VersionUid,
                       volume: str,
                       snapshot: str,
                       block_size: int = None,
                       storage_name: str = None,
                       size: int = None,
                       base_version_uid: VersionUid = None,
                       base_version_locking: bool = True) -> Version:
        """ Prepares the metadata for a new version.
        If base_version_uid is given, this is taken as the base, otherwise
        a pure sparse version is created.
        """
        if storage_name:
            new_storage = Storage.get_by_name(storage_name)
        else:
            new_storage = Storage.get_by_name(self._default_storage_name)

        old_blocks_iter: Optional[Iterator[Block]] = None
        if base_version_uid:
            if not base_version_locking and not Locking.is_version_locked(base_version_uid):
                raise InternalError('Base version is not locked.')
            old_version = Version.get_by_uid(base_version_uid)  # raise if not exists
            if not old_version.status.is_valid():
                raise UsageError('You can only base a new version on a valid old version.')
            if block_size is not None and old_version.block_size != block_size:
                raise UsageError('You cannot base a new version on an old version with a different block size.')
            if old_version.storage != new_storage:
                raise UsageError(f'Base version and new version have to be in the same storage ({old_version.storage.name} != {new_storage.name}).')

            base_version = Version.get_by_uid(base_version_uid)
            old_blocks_iter = base_version.blocks

            if size is not None:
                new_size = size
            else:
                new_size = old_version.size

            new_version_block_size = old_version.block_size
        else:
            if size is None:
                raise InternalError('Size needs to be specified if there is no base version.')
            new_size = size
            if block_size is None:
                raise InternalError('Block size needs to be specified if there is no base version.')
            new_version_block_size = block_size

        version = None
        try:
            if base_version_locking and base_version_uid:
                Locking.lock_version(base_version_uid, reason='Base version cloning')

            # We always start with invalid versions, then mark them valid after the backup succeeds.
            version = Version.create(version_uid=version_uid,
                                     volume=volume,
                                     snapshot=snapshot,
                                     size=new_size,
                                     block_size=new_version_block_size,
                                     storage_id=new_storage.id,
                                     status=VersionStatus.incomplete)
            Locking.lock_version(version.uid, reason='Preparing version')

            new_block_uid: Optional[BlockUid]
            new_block_checksum: Optional[str]
            new_block_block_size: int
            new_block_valid: bool
            blocks: List[Dict[str, Any]] = []
            old_version_exhausted = old_blocks_iter is None
            for idx in range(version.blocks_count):
                if not old_version_exhausted:
                    assert old_blocks_iter is not None
                    old_block = next(old_blocks_iter, None)
                    if old_block is not None:
                        assert old_block.idx == idx
                        new_block_uid = old_block.uid
                        new_block_checksum = old_block.checksum
                        new_block_block_size = old_block.size
                        new_block_valid = old_block.valid
                    else:
                        old_version_exhausted = True
                        new_block_uid = None
                        new_block_checksum = None
                        new_block_block_size = version.block_size
                        new_block_valid = True
                else:
                    new_block_uid = None
                    new_block_checksum = None
                    new_block_block_size = version.block_size
                    new_block_valid = True

                # This catches blocks which changed their size between the base version and the new version.
                # If the new version is bigger, this affects the last block of the old version and the last block of
                # the new version.
                # If the new version is smaller, this affects the last block of the new version.
                offset = idx * version.block_size
                new_block_block_size_tmp = min(version.block_size, new_size - offset)
                if new_block_block_size != new_block_block_size_tmp:
                    new_block_block_size = new_block_block_size_tmp
                    new_block_uid = None
                    new_block_checksum = None
                    # Forces reread.
                    new_block_valid = False

                blocks.append({
                    'idx': idx,
                    'uid_left': new_block_uid.left if new_block_uid is not None else None,
                    'uid_right': new_block_uid.right if new_block_uid is not None else None,
                    'checksum': new_block_checksum,
                    'size': new_block_block_size,
                    'valid': new_block_valid
                })

                notify(self._process_name,
                       'Preparing version {} ({:.1f}%)'.format(version.uid, (idx + 1) / version.blocks_count * 100))

                if len(blocks) == self._BLOCKS_CREATE_WORK_PACKAGE:
                    version.create_blocks(blocks=blocks)
                    blocks = []
            version.create_blocks(blocks=blocks)
        except:
            if version and Locking.is_version_locked(version.uid):
                Locking.unlock_version(version.uid)
            raise
        finally:
            if base_version_locking and base_version_uid and Locking.is_version_locked(base_version_uid):
                Locking.unlock_version(base_version_uid)
            notify(self._process_name)

        return version

    @staticmethod
    def get_version_by_uid(version_uid: VersionUid) -> Version:
        return Version.get_by_uid(version_uid)

    @staticmethod
    def find_versions_with_filter(filter_expression: str = None) -> List[Version]:
        return Version.find_with_filter(filter_expression)

    def _scrub_prepare(self,
                       *,
                       version: Version,
                       history: BlockUidHistory = None,
                       block_percentage: int,
                       deep_scrub: bool) -> int:
        storage = StorageFactory.get_by_name(version.storage.name)
        read_jobs = 0
        for i, block in enumerate(version.blocks):
            notify(
                self._process_name,
                'Preparing {} of version {} ({:.1f}%)'.format('deep-scrub' if deep_scrub else 'scrub', version.uid,
                                                              (i + 1) / version.blocks_count * 100))
            if not block.uid:
                logger.debug('{} of block {} of version {} (UID {}) skipped (sparse).'.format(
                    'Deep-scrub' if deep_scrub else 'Scrub', block.idx, version.uid, block.uid))
                continue
            if history and history.seen(version.storage_id, block.uid):
                logger.debug('{} of block {} of version {} (UID {}) skipped (already seen).'.format(
                    'Deep-scrub' if deep_scrub else 'Scrub', block.idx, version.uid, block.uid))
                continue
            # i != 0 ensures that we always scrub at least one block (the first in this case)
            if i != 0 and block_percentage < 100 and random.randint(1, 100) > block_percentage:
                logger.debug('{} of block {} of version {} (UID {}) skipped (percentile is {}).'.format(
                    'Deep-scrub' if deep_scrub else 'Scrub', block.idx, version.uid, block.uid, block_percentage))
            else:
                storage.read_block_async(block, metadata_only=(not deep_scrub))
                read_jobs += 1
        return read_jobs

    def _scrub_report_progress(self, *, version_uid: VersionUid, block: DereferencedBlock, read_jobs: int,
                               done_read_jobs: int, deep_scrub: bool) -> None:
        logger.debug('{} of block {} (UID {}) ok.'.format('Deep-scrub' if deep_scrub else 'Scrub', block.idx, block.uid))

        notify(
            self._process_name, '{} version {} ({:.1f}%)'.format('Deep-scrubbing' if deep_scrub else 'Scrubbing',
                                                                 version_uid, done_read_jobs / read_jobs * 100))
        log_every_jobs = read_jobs // 200 + 1  # about every half percent
        if done_read_jobs % log_every_jobs == 0 or done_read_jobs == read_jobs:
            logger.info('{} {}/{} blocks ({:.1f}%)'.format('Deep-scrubbed' if deep_scrub else 'Scrubbed',
                                                           done_read_jobs, read_jobs, done_read_jobs / read_jobs * 100))

    def scrub(self, version_uid: VersionUid, block_percentage: int = 100, history: BlockUidHistory = None) -> None:
        Locking.lock_version(version_uid, reason='Scrubbing version')
        try:
            version = Version.get_by_uid(version_uid)
            if not version.status.is_scrubbable():
                raise ScrubbingError('Version {} cannot be scrubbed, it has a status of {}.'.format(
                    version_uid, version.status.name))

        except:
            Locking.unlock_version(version_uid)
            raise

        valid = True
        affected_version_uids = []
        try:
            storage = StorageFactory.get_by_name(version.storage.name)
            read_jobs = self._scrub_prepare(version=version,
                                            history=history,
                                            block_percentage=block_percentage,
                                            deep_scrub=False)

            done_read_jobs = 0
            for entry in storage.read_get_completed():
                done_read_jobs += 1
                if isinstance(entry, Exception):
                    # If it really is a data inconsistency mark blocks invalid
                    if isinstance(entry, InvalidBlockException):
                        logger.error('Block {} (UID {}) is invalid: {}{}'.format(
                            entry.block.idx, entry.block.uid, entry,
                            f' Caused by: {entry.__cause__}' if entry.__cause__ else ''))
                        affected_version_uids.extend(Version.set_block_valid(entry.block.uid, False))
                        valid = False
                        continue
                    else:
                        raise entry
                else:
                    block, data, metadata = cast(Tuple[DereferencedBlock, bytes, Dict], entry)

                try:
                    storage.check_block_metadata(block=block, data_length=None, metadata=metadata)
                except (KeyError, ValueError) as exception:
                    logger.error('Metadata check failed, block {} (UID {}) is invalid: {}'.format(
                        block.idx, block.uid, exception))
                    affected_version_uids.extend(Version.set_block_valid(block.uid, False))
                    valid = False
                    continue

                if history:
                    history.add(version.storage_id, block.uid)

                self._scrub_report_progress(version_uid=version_uid,
                                            block=block,
                                            read_jobs=read_jobs,
                                            done_read_jobs=done_read_jobs,
                                            deep_scrub=False)
        finally:
            Locking.unlock_version(version_uid)
            notify(self._process_name)

        if read_jobs != done_read_jobs:
            raise InternalError(
                'Number of submitted and completed read jobs inconsistent (submitted: {}, completed {}).'.format(
                    read_jobs, done_read_jobs))

        # A scrub (in contrast to a deep-scrub) can only ever mark a version as invalid. To mark it as valid
        # there is not enough information.
        if valid:
            logger.info('Scrub of version {} successful.'.format(version.uid))
        else:
            logger.error('Marked version {} as invalid because it has errors.'.format(version_uid))
            affected_version_uids.remove(version_uid)
            if affected_version_uids:
                logger.error('Marked the following versions as invalid, too, because of invalid blocks: {}.'\
                             .format(', '.join(sorted(affected_version_uids))))
            raise ScrubbingError('Scrub of version {} failed.'.format(version_uid))

    def deep_scrub(self,
                   version_uid: VersionUid,
                   source: str = None,
                   block_percentage: int = 100,
                   history: BlockUidHistory = None) -> None:
        Locking.lock_version(version_uid, reason='Deep-scrubbing')
        try:
            version = Version.get_by_uid(version_uid)
            if not version.status.is_deep_scrubbable():
                raise ScrubbingError('Version {} cannot be deep-scrubbed, it has a status of {}.'.format(
                    version_uid, version.status.name))
            # This only outputs a warning to let the user know that the version it not valid to begin with.
            if not version.status.is_valid():
                logger.warn('Version {} has a status of {}.'.format(version_uid, version.status.name))

            if source:
                io = IOFactory.get(source, version.block_size)
                io.open_r()
        except:
            Locking.unlock_version(version_uid)
            raise

        valid = True
        source_mismatch = False
        affected_version_uids = []
        try:
            storage = StorageFactory.get_by_name(version.storage.name)
            old_use_read_cache = storage.use_read_cache(False)
            read_jobs = self._scrub_prepare(version=version,
                                            history=history,
                                            block_percentage=block_percentage,
                                            deep_scrub=True)

            done_read_jobs = 0
            for entry in storage.read_get_completed():
                done_read_jobs += 1
                if isinstance(entry, Exception):
                    # If it really is a data inconsistency mark blocks invalid
                    if isinstance(entry, InvalidBlockException):
                        logger.error('Block {} (UID {}) is invalid: {}{}'.format(
                            entry.block.idx, entry.block.uid, entry,
                            f' Caused by: {entry.__cause__}' if entry.__cause__ else ''))
                        affected_version_uids.extend(Version.set_block_valid(entry.block.uid, False))
                        valid = False
                        continue
                    else:
                        raise entry
                else:
                    block, data, metadata = cast(Tuple[DereferencedBlock, bytes, Dict], entry)

                try:
                    storage.check_block_metadata(block=block, data_length=len(data), metadata=metadata)
                except (KeyError, ValueError) as exception:
                    logger.error('Metadata check failed, block {} of version {} (UID {}) is invalid: {}'.format(
                        block.idx, version_uid, block.uid, exception))
                    Version.set_block_valid(block.uid, False)
                    valid = False
                    continue

                data_checksum = self._block_hash.data_hexdigest(data)
                if data_checksum != block.checksum:
                    logger.error(
                        'Checksum mismatch during deep-scrub of block {} of version {} (UID {}) (is: {}... should-be: {}...).'.format(
                            block.idx, version_uid, block.uid, data_checksum[:16],
                            cast(str, block.checksum)[:16]))  # We know that block.checksum is set
                    affected_version_uids.extend(Version.set_block_valid(block.uid, False))
                    valid = False
                    continue

                if source:
                    source_data = io.read_sync(block)
                    if source_data != data:
                        logger.error('Source data has changed for block {} of version {} (UID {}) (is: {}... should-be: {}...). '
                                     'Will not mark this block as invalid, because the source looks wrong.'.format(
                                         block.idx, version_uid, block.uid,
                                         self._block_hash.data_hexdigest(source_data)[:16], data_checksum[:16]))
                        valid = False
                        # We are not setting the block invalid here because when the block exists and the checksum is
                        # correct then the source is probably invalid.
                        source_mismatch = True

                if not block.valid:
                    logger.info('Block {} of version {} (UID {}) passed revalidation, marking it as valid.'.format(
                        block.idx, version_uid, block.uid))
                    Version.set_block_valid(block.uid, True)

                # Only add the block to the history if it is valid. This ensure that this block will also get flagged
                # again when deep-scrubbing other versions containing it.
                if history:
                    history.add(version.storage_id, block.uid)

                self._scrub_report_progress(version_uid=version_uid,
                                            block=block,
                                            read_jobs=read_jobs,
                                            done_read_jobs=done_read_jobs,
                                            deep_scrub=True)
        except:
            Locking.unlock_version(version_uid)
            raise
        finally:
            if source:
                io.close()
            # Restore old read cache setting
            storage.use_read_cache(old_use_read_cache)
            notify(self._process_name)

        if read_jobs != done_read_jobs:
            raise InternalError(
                'Number of submitted and completed read jobs inconsistent (submitted: {}, completed {}).'.format(
                    read_jobs, done_read_jobs))

        if valid:
            if block_percentage == 100:
                try:
                    version.set(status=VersionStatus.valid)
                except:
                    Locking.unlock_version(version_uid)
                    raise
            logger.info('Deep-scrub of version {} successful.'.format(version.uid))
        else:
            if source_mismatch:
                logger.error('Version {} had source mismatches.'.format(version.uid))
            logger.error('Marked version {} as invalid because it has errors.'.format(version.uid))
            if version.uid in affected_version_uids:
                affected_version_uids.remove(version.uid)
            if affected_version_uids:
                logger.error('Marked the following versions as invalid too because of invalid blocks: {}.' \
                             .format(', '.join(sorted(affected_version_uids))))

        Locking.unlock_version(version_uid)

        if not valid:
            raise ScrubbingError('Deep-scrub of version {} failed.'.format(version_uid))

    def _batch_scrub(self, method: str, filter_expression: Optional[str], version_percentage: int,
                     block_percentage: int, group_label: Optional[str]) -> Tuple[List[Version], List[Version]]:
        history = BlockUidHistory()
        versions = set(Version.find_with_filter(filter_expression))
        errors = []

        if versions and group_label is not None:
            additional_versions: Set[Version] = set()
            for version in versions:
                label_match = list(filter(lambda label: label.name == group_label, version.labels))
                if not label_match:
                    continue
                assert len(label_match) == 1
                additional_versions |= set(Version.find(labels=[(group_label, label_match[0].value)]))
            versions |= additional_versions

        if version_percentage and versions:
            # Will always scrub at least one matching version
            versions = set(random.sample(list(versions), max(1, int(len(versions) * version_percentage / 100))))
        if not versions:
            logger.info('No matching versions found.')
            return [], []

        for version in versions:
            try:
                logger.info('{} {}% of version {} (volume {}).'.format(
                    'Scrubbing' if method == 'scrub' else 'Deep-scrubbing', block_percentage, version.uid,
                    version.volume))
                getattr(self, method)(version.uid, block_percentage=block_percentage, history=history)
            except ScrubbingError as exception:
                logger.error(str(exception))
                errors.append(version)
            except AlreadyLocked:
                logger.warning(f'Skipping version {version.uid}, it is locked. ')

        return sorted(versions), sorted(errors)

    def batch_scrub(self,
                    filter_expression: Optional[str],
                    version_percentage: int,
                    block_percentage: int,
                    group_label: Optional[str] = None) -> Tuple[List[Version], List[Version]]:
        return self._batch_scrub('scrub', filter_expression, version_percentage, block_percentage, group_label)

    def batch_deep_scrub(self,
                         filter_expression: Optional[str],
                         version_percentage: int,
                         block_percentage: int,
                         group_label: Optional[str] = None) -> Tuple[List[Version], List[Version]]:
        return self._batch_scrub('deep_scrub', filter_expression, version_percentage, block_percentage, group_label)

    def restore(self, version_uid: VersionUid, target: str, sparse: bool = False, force: bool = False) -> None:
        block: Union[DereferencedBlock, Block]

        Locking.lock_version(version_uid, reason='Restoring version')
        try:
            version = Version.get_by_uid(version_uid)  # raise if version not exists
            notify(self._process_name, 'Restoring version {} to {}: Getting blocks'.format(version_uid, target))

            self._storage = version.storage_id

            io = IOFactory.get(target, version.block_size)
            io.open_w(version.size, force=force, sparse=sparse)
        except:
            Locking.unlock_version(version_uid)
            raise

        try:
            t1 = time.time()
            storage = StorageFactory.get_by_name(version.storage.name)

            sparse_blocks_count = version.sparse_blocks_count
            read_blocks_count = version.blocks_count - sparse_blocks_count

            read_jobs = 0
            write_jobs = 0
            done_write_jobs = 0
            written = 0
            log_every_jobs = (sparse_blocks_count if not sparse else read_blocks_count) // 200 + 1  # about every half percent
            sparse_data_block = b'\0' * version.block_size

            def handle_sparse_write_completed(timeout: int = None):
                nonlocal done_write_jobs, written, sparse_blocks_count, log_every_jobs, version_uid, target
                try:
                    for written_block in io.write_get_completed(timeout=timeout):
                        if isinstance(written_block, Exception):
                            raise written_block
                        assert isinstance(written_block, DereferencedBlock)
                        done_write_jobs += 1
                        written += written_block.size

                        notify(
                            self._process_name, 'Restoring version {} to {}: Sparse writing ({:.1f}%)'.format(
                                version_uid, target, done_write_jobs / sparse_blocks_count * 100))
                        if done_write_jobs % log_every_jobs == 0 or done_write_jobs == sparse_blocks_count:
                            logger.info('Wrote sparse {}/{} blocks ({:.1f}%)'.format(
                                done_write_jobs, sparse_blocks_count, done_write_jobs / sparse_blocks_count * 100))
                except (TimeoutError, CancelledError):
                    pass

            for block in version.blocks:
                if block.uid:
                    storage.read_block_async(block)
                    read_jobs += 1
                    logger.debug('Queued read for block {} successfully ({} bytes).'.format(block.idx, block.size))
                elif not sparse:
                    io.write(block, sparse_data_block)
                    write_jobs += 1
                    logger.debug('Queued write for sparse block {} successfully ({} bytes).'.format(
                        block.idx, block.size))
                else:
                    logger.debug('Ignored sparse block {}.'.format(block.idx))

                if sparse:
                    # Version might be completely sparse
                    if read_blocks_count > 0:
                        notify(
                            self._process_name, 'Restoring version {} to {}: Queueing blocks to read ({:.1f}%)'.format(
                                version_uid, target, read_jobs / read_blocks_count * 100))
                else:
                    handle_sparse_write_completed(timeout=0)

            handle_sparse_write_completed()

            if write_jobs != done_write_jobs:
                raise InternalError(
                    'Number of submitted and completed write jobs inconsistent (submitted: {}, completed {}).'.format(
                        write_jobs, done_write_jobs))

            done_read_jobs = 0
            write_jobs = 0
            done_write_jobs = 0
            log_every_jobs = read_jobs // 200 + 1  # about every half percent

            def handle_write_completed(timeout: int = None):
                nonlocal done_write_jobs, written, read_jobs, log_every_jobs, version_uid, target
                try:
                    for written_block in io.write_get_completed(timeout=timeout):
                        if isinstance(written_block, Exception):
                            raise written_block
                        assert isinstance(written_block, DereferencedBlock)
                        done_write_jobs += 1
                        written += written_block.size

                        notify(
                            self._process_name,
                            'Restoring version {} to {} ({:.1f}%)'.format(version_uid, target,
                                                                          done_write_jobs / read_jobs * 100))
                        if done_write_jobs % log_every_jobs == 0 or done_write_jobs == read_jobs:
                            logger.info('Restored {}/{} blocks ({:.1f}%)'.format(done_write_jobs, read_jobs,
                                                                                 done_write_jobs / read_jobs * 100))
                except (TimeoutError, CancelledError):
                    pass

            for entry in storage.read_get_completed():
                done_read_jobs += 1
                if isinstance(entry, Exception):
                    logger.error('Storage backend read failed: {}'.format(entry))
                    # If it really is a data inconsistency mark blocks invalid
                    if isinstance(entry, (KeyError, ValueError)):
                        Version.set_block_valid(block.uid, False)
                        continue
                    else:
                        raise entry
                else:
                    block, data, metadata = cast(Tuple[DereferencedBlock, bytes, Dict], entry)

                # Write what we have
                io.write(block, data)
                write_jobs += 1

                try:
                    storage.check_block_metadata(block=block, data_length=len(data), metadata=metadata)
                except (KeyError, ValueError) as exception:
                    logger.error('Metadata check failed, block is invalid: {}'.format(exception))
                    Version.set_block_valid(block.uid, False)
                    continue

                data_checksum = self._block_hash.data_hexdigest(data)
                if data_checksum != block.checksum:
                    logger.error('Checksum mismatch during restore for block {} (UID {}) (is: {}... should-be: {}..., '
                                 'block.valid: {}). Block restored is invalid.'.format(
                                     block.idx, block.uid, data_checksum[:16],
                                     cast(str, block.checksum)[:16], block.valid))  # We know that block.checksum is set
                    Version.set_block_valid(block.uid, False)
                else:
                    logger.debug('Restored block {} successfully ({} bytes).'.format(block.idx, block.size))

                handle_write_completed(timeout=0)

            handle_write_completed()
        finally:
            io.close()
            t2 = time.time()
            Locking.unlock_version(version_uid)
            notify(self._process_name)

        if read_jobs != done_read_jobs:
            raise InternalError(
                'Number of submitted and completed read jobs inconsistent (submitted: {}, completed {}).'.format(
                    read_jobs, done_read_jobs))

        if write_jobs != done_write_jobs:
            raise InternalError(
                'Number of submitted and completed write jobs inconsistent (submitted: {}, completed {}).'.format(
                    write_jobs, done_write_jobs))

        logger.info('Successfully restored version {} in {} with {}/s.'.format(
            version.uid, PrettyPrint.duration(max(int(t2 - t1), 1)), PrettyPrint.bytes(written / (t2 - t1))))

    @staticmethod
    def protect(version_uid: VersionUid, protected: bool) -> None:
        version = Version.get_by_uid(version_uid)
        version.set(protected=protected)

    @staticmethod
    def rm(version_uid: VersionUid,
           force: bool = True,
           disallow_rm_when_younger_than_days: int = 0,
           keep_metadata_backup: bool = False,
           override_lock: bool = False) -> None:
        with Locking.with_version_lock(version_uid, reason='Removing version', override_lock=override_lock):
            version = Version.get_by_uid(version_uid)

            if version.protected:
                raise PermissionError('Version {} is protected, will not delete it.'.format(version_uid))

            if not force:
                # check if disallow_rm_when_younger_than_days allows deletion
                age_days = (datetime.datetime.now() - version.date).days
                if disallow_rm_when_younger_than_days > age_days:
                    raise PermissionError('Version {} is too young. Will not delete.'.format(version_uid))
                if not version.status.is_removable():
                    raise PermissionError('Version {} cannot be removed without force, it has status {}.'.format(
                        version_uid, version.status.name))

            num_blocks = version.remove()

            if not keep_metadata_backup:
                try:
                    storage = StorageFactory.get_by_name(version.storage.name)
                    storage.rm_version(version_uid)
                    logger.info('Removed version {} metadata backup from storage.'.format(version_uid))
                except FileNotFoundError:
                    logger.warning(
                        'Unable to remove version {} metadata backup from storage, the object was not found.'.format(version_uid))
                    pass

            logger.info('Removed backup version {} with {} blocks.'.format(version_uid, num_blocks))

    @staticmethod
    def _blocks_from_hints(hints: Sequence[Tuple[int, int, bool]], block_size: int) -> Tuple[Set[int], Set[int]]:
        sparse_blocks = set()
        read_blocks = set()
        for offset, length, exists in hints:
            start_block = offset // block_size
            end_block = (offset + length - 1) // block_size
            if exists:
                for i in range(start_block, end_block + 1):
                    read_blocks.add(i)
            else:
                if offset % block_size > 0:
                    # Start block is only partially sparse, make sure it is read
                    read_blocks.add(start_block)

                if (offset + length) % block_size > 0:
                    # End block is only partially sparse, make sure it is read
                    read_blocks.add(end_block)

                for i in range(start_block, end_block + 1):
                    sparse_blocks.add(i)

        return sparse_blocks, read_blocks

    def backup(self,
               *,
               version_uid: VersionUid,
               volume: str,
               snapshot: str,
               source: str,
               hints: List[Tuple[int, int, bool]] = None,
               base_version_uid: VersionUid = None,
               storage_name: str = None,
               block_size: int = None) -> Version:
        """ Create a backup from source.
        If hints are given, they must be tuples of (offset, length, exists) where offset and length are integers and
        exists is a boolean. In this case only data within hints will be backed up.
        Otherwise, the backup reads source and looks if checksums match with the target.
        """
        if not InputValidation.is_volume_name(volume):
            raise UsageError('Version name {} is invalid.'.format(volume))
        if not InputValidation.is_snapshot_name(snapshot):
            raise UsageError('Snapshot name {} is invalid.'.format(snapshot))

        block: Union[DereferencedBlock, Block]

        stats: Dict[str, Any] = {
            'bytes_read': 0,
            'bytes_written': 0,
            'bytes_deduplicated': 0,
            'bytes_sparse': 0,
            'start_time': time.time(),
        }

        new_version_block_size = block_size if block_size else self._block_size

        io = IOFactory.get(source, block_size=new_version_block_size)
        io.open_r()

        source_size = io.size()

        version = self.create_version(version_uid=version_uid,
                                      volume=volume,
                                      snapshot=snapshot,
                                      size=source_size,
                                      block_size=new_version_block_size,
                                      base_version_uid=base_version_uid,
                                      storage_name=storage_name)
        Locking.update_version_lock(version.uid, reason='Backing up')

        if hints is not None:
            if len(hints) > 0:
                # Sanity check: check hints for validity, i.e. too high offsets, ...
                max_offset = max(h[0] + h[1] for h in hints)
                if max_offset > source_size:
                    raise InputDataError('Hints have higher offsets than source file.')

                sparse_blocks, read_blocks = self._blocks_from_hints(hints, version.block_size)
            else:
                # Two snapshots can be completely identical between one backup and next
                logger.warning('Hints are empty, assuming nothing has changed.')
                sparse_blocks = set()
                read_blocks = set()
        else:
            sparse_blocks = set()
            read_blocks = set(range(version.blocks_count))

        if base_version_uid and hints is not None:
            # SANITY CHECK:
            # Check some blocks outside of hints if they are the same in the
            # base_version backup and in the current backup. If they
            # aren't, either hints are wrong (e.g. from a wrong snapshot diff)
            # or source doesn't match. In any case, the resulting backup won't
            # be good.
            logger.info('Starting sanity check with 0.1% of the ignored blocks.')
            notify(self._process_name, 'Sanity checking hints of version {}'.format(version.uid))

            ignored_blocks = sorted(set(range(version.blocks_count)) - read_blocks - sparse_blocks)
            # 0.1% but at least ten. If there are less than ten blocks check them all.
            check_blocks_count = max(min(len(ignored_blocks), 10), len(ignored_blocks) // 1000)
            # 50% from the start
            check_blocks = set(ignored_blocks[:check_blocks_count // 2])
            # and 50% from random locations
            check_blocks = check_blocks.union(random.sample(ignored_blocks, check_blocks_count // 2))
            read_jobs = 0
            for block in [version.get_block_by_idx(idx) for idx in check_blocks]:
                if block.uid and block.valid:  # no uid = sparse block in backup. Can't check.
                    io.read(block)
                    read_jobs += 1

            for entry in io.read_get_completed():
                if isinstance(entry, Exception):
                    raise entry
                else:
                    source_block, source_data = cast(Tuple[DereferencedBlock, bytes], entry)

                # check metadata checksum with the newly read one
                source_data_checksum = self._block_hash.data_hexdigest(source_data)
                if source_block.checksum != source_data_checksum:
                    logger.error("Source and backup don't match in regions outside of the ones indicated by the hints.")
                    logger.error("Looks like the hints don't match or the source is different.")
                    logger.error("Found wrong source data at block {}: offset {}, length {}".format(
                        source_block.idx, source_block.idx * version.block_size, version.block_size))
                    # remove version
                    version.remove()
                    Locking.unlock_version(version.uid)
                    raise InputDataError('Source changed in regions outside of ones indicated by the hints.')
            logger.info('Finished sanity check. Checked {} blocks.'.format(read_jobs))

        logger.info(f'Starting backup of {io.sanitized_url} to storage {version.storage.name}, the newly created version is {version.uid}.')
        try:
            storage = StorageFactory.get_by_name(version.storage.name)
            read_jobs = 0
            for block in version.blocks:
                if block.idx in read_blocks or not block.valid:
                    io.read(block)
                    read_jobs += 1
                elif block.idx in sparse_blocks:
                    # This "elif" is very important. Because if the block is in read_blocks AND sparse_blocks,
                    # it *must* be read.

                    # Only update the database when the block wasn't sparse to begin with
                    if block.uid:
                        version.set_block(idx=block.idx,
                                          block_uid=SparseBlockUid,
                                          checksum=None,
                                          size=block.size,
                                          valid=True)
                        logger.debug('Skipping block (had data, turned sparse) {}'.format(block.idx))
                    else:
                        assert block.checksum is None
                        logger.debug('Skipping block (sparse) {}'.format(block.idx))
                    stats['bytes_sparse'] += block.size

                else:
                    # Block is already in database, no need to update it
                    logger.debug('Keeping block {}'.format(block.idx))
                notify(
                    self._process_name, 'Backing up version {} from {}: Queueing blocks to read ({:.1f}%)'.format(
                        version.uid, source, (block.idx + 1) / version.blocks_count * 100))

            # Precompute checksum of a sparse block.
            sparse_block_checksum = self._block_hash.data_hexdigest(b'\0' * version.block_size)

            done_read_jobs = 0
            write_jobs = 0
            done_write_jobs = 0
            log_every_jobs = read_jobs // 200 + 1  # about every half percent

            def handle_write_completed(timeout: int = None):
                nonlocal done_write_jobs, stats, version
                try:
                    for written_block in storage.write_get_completed(timeout=timeout):
                        if isinstance(written_block, Exception):
                            raise written_block

                        written_block = cast(DereferencedBlock, written_block)

                        assert written_block.version_id == version.id
                        version.set_block(idx=written_block.idx,
                                          block_uid=written_block.uid,
                                          checksum=written_block.checksum,
                                          size=written_block.size,
                                          valid=True)
                        done_write_jobs += 1
                        stats['bytes_written'] += written_block.size
                except (TimeoutError, CancelledError):
                    pass

            for entry in io.read_get_completed():
                if isinstance(entry, Exception):
                    raise entry
                else:
                    block, data = cast(Tuple[DereferencedBlock, bytes], entry)

                stats['bytes_read'] += len(data)

                data_checksum = self._block_hash.data_hexdigest(data)
                if data_checksum == sparse_block_checksum and block.size == version.block_size:
                    # It's a sparse block.
                    stats['bytes_sparse'] += block.size
                    logger.debug('Skipping block (detected sparse) {}'.format(block.idx))
                    version.set_block(idx=block.idx,
                                      block_uid=SparseBlockUid,
                                      checksum=None,
                                      size=block.size,
                                      valid=True)
                else:
                    existing_block = version.get_block_by_checksum(data_checksum)
                    if existing_block and existing_block.size == block.size:
                        # It's a known block.
                        version.set_block(idx=block.idx,
                                          block_uid=existing_block.uid,
                                          checksum=existing_block.checksum,
                                          size=existing_block.size,
                                          valid=True)
                        stats['bytes_deduplicated'] += len(data)
                        logger.debug('Found existing block for id {} with UID {}'.format(block.idx, existing_block.uid))
                    else:
                        # It's a new block.
                        # Generate a unique block id by combining the version id and the block index.
                        block.uid = BlockUid(version.id, block.idx + 1)
                        block.checksum = data_checksum
                        storage.write_block_async(block, data)
                        write_jobs += 1
                        logger.debug('Queued block {} for write (checksum {}...)'.format(block.idx, data_checksum[:16]))

                done_read_jobs += 1

                handle_write_completed(timeout=0)

                notify(
                    self._process_name,
                    'Backing up version {} from {} ({:.1f}%)'.format(version.uid, source,
                                                                     done_read_jobs / read_jobs * 100))
                if done_read_jobs % log_every_jobs == 0 or done_read_jobs == read_jobs:
                    logger.info('Backed up {}/{} blocks ({:.1f}%)'.format(done_read_jobs, read_jobs,
                                                                          done_read_jobs / read_jobs * 100))

            handle_write_completed()
        except:
            Locking.unlock_version(version.uid)
            raise
        finally:
            # This will also cancel any outstanding read jobs
            io.close()
            version.commit()

        if read_jobs != done_read_jobs:
            raise InternalError(
                'Number of submitted and completed read jobs inconsistent (submitted: {}, completed {}).'.format(
                    read_jobs, done_read_jobs))

        if write_jobs != done_write_jobs:
            raise InternalError(
                'Number of submitted and completed write jobs inconsistent (submitted: {}, completed {}).'.format(
                    write_jobs, done_write_jobs))

        notify(self._process_name, 'Marking version {} as valid'.format(version.uid))
        version.set(status=VersionStatus.valid)

        notify(self._process_name, 'Backing up metadata of version {}'.format(version.uid))
        self.metadata_backup([version.uid], overwrite=True, locking=False)

        logger.debug('Stats: {}'.format(stats))
        version.set_stats(
            bytes_read=stats['bytes_read'],
            bytes_written=stats['bytes_written'],
            bytes_deduplicated=stats['bytes_deduplicated'],
            bytes_sparse=stats['bytes_sparse'],
            duration=int(time.time() - stats['start_time']),
        )

        Locking.unlock_version(version.uid)
        notify(self._process_name)
        logger.info('Backup successful.'.format(version.uid))
        return version

    def cleanup(self, dt: int = 3600, override_lock: bool = False) -> None:
        with Locking.with_lock(lock_name='cleanup',
                               reason='Cleanup',
                               locked_msg='Another cleanup is already running.',
                               override_lock=override_lock):
            notify(self._process_name, 'Cleanup')
            for hit_list in DeletedBlock.get_unused_block_uids(dt):
                for storage_name, uids in hit_list.items():
                    storage = StorageFactory.get_by_name(storage_name)
                    logger.debug('Deleting UIDs from storage {}: {}'.format(storage_name,
                                                                            ', '.join(str(uid) for uid in uids)))

                    for uid in uids:
                        storage.rm_block_async(uid)

                    no_del_uids = []
                    for entry in storage.rm_get_completed():
                        if isinstance(entry, BlockNotFoundError):
                            no_del_uids.append(entry.uid)
                        elif isinstance(entry, Exception):
                            raise entry

                    if no_del_uids:
                        logger.info('Unable to delete these UIDs from storage {}: {}'.format(
                            storage_name, ', '.join(str(uid) for uid in no_del_uids)))
            notify(self._process_name)

    @staticmethod
    def add_label(version_uid: VersionUid, key: str, value: str) -> None:
        version = Version.get_by_uid(version_uid)
        version.add_label(key, value)

    @staticmethod
    def rm_label(version_uid: VersionUid, key: str) -> None:
        version = Version.get_by_uid(version_uid)
        version.rm_label(key)

    @staticmethod
    def close() -> None:
        StorageFactory.close()
        # Close database backend after storage so that any open locks are held until all storage jobs have
        # finished.
        Database.close()

    @staticmethod
    def metadata_export(version_uids: Sequence[VersionUid], f: TextIO) -> None:
        try:
            locked_version_uids = []
            for version_uid in version_uids:
                Locking.lock_version(version_uid, reason='Exporting version metadata')
                locked_version_uids.append(version_uid)

            Database.export(version_uids, f)
            logger.info('Exported metadata of version(s): {}.'.format(', '.join(version_uids)))
        finally:
            for version_uid in locked_version_uids:
                Locking.unlock_version(version_uid)

    @staticmethod
    def metadata_backup(version_uids: Sequence[VersionUid], overwrite: bool = False, locking: bool = True) -> None:
        versions = [Version.get_by_uid(version_uid) for version_uid in version_uids]
        try:
            locked_version_uids = []
            if locking:
                for version in versions:
                    Locking.lock_version(version.uid, reason='Backing up version metadata')
                    locked_version_uids.append(version.uid)

            for version in versions:
                with StringIO() as metadata_export:
                    Database.export([version.uid], metadata_export)
                    storage = StorageFactory.get_by_name(version.storage.name)
                    logger.debug(metadata_export.getvalue())
                    storage.write_version(version.uid, metadata_export.getvalue(), overwrite=overwrite)
                logger.info('Backed up metadata of version {}.'.format(version.uid))
        finally:
            for version_uid in locked_version_uids:
                Locking.unlock_version(version_uid)

    @staticmethod
    def export_any(*args, **kwargs) -> None:
        Database.export_any(*args, **kwargs)

    @staticmethod
    def metadata_import(f: TextIO) -> None:
        # TODO: Find a good way to lock here
        version_uids = Database.import_(f)
        logger.info('Imported metadata of version(s): {}.'.format(', '.join(version_uids)))

    def metadata_restore(self, version_uids: Sequence[VersionUid], storage_name: str = None) -> None:
        storage = StorageFactory.get_by_name(storage_name or self._default_storage_name)
        try:
            locked_version_uids = []
            for version_uid in version_uids:
                Locking.lock_version(version_uid, reason='Restoring version metadata')
                locked_version_uids.append(version_uid)

            for version_uid in version_uids:
                metadata_import_data = storage.read_version(version_uid)
                with StringIO(metadata_import_data) as metadata_import:
                    Database.import_(metadata_import)
                logger.info('Restored metadata of version {}.'.format(version_uid))
        finally:
            for version_uid in locked_version_uids:
                Locking.unlock_version(version_uid)

    def metadata_ls(self, storage_name: str = None) -> List[VersionUid]:
        storage = StorageFactory.get_by_name(storage_name or self._default_storage_name)
        return storage.list_versions()

    def enforce_retention_policy(self,
                                 filter_expression: str,
                                 rules_spec: str,
                                 dry_run: bool = False,
                                 keep_metadata_backup: bool = False,
                                 group_label: str = None) -> List[Version]:
        versions = Version.find_with_filter(filter_expression)

        versions_by_volume: Dict[str, List[Version]] = defaultdict(list)
        for version in versions:
            if version.protected:
                logger.info('Not considering version {}, it is protected.'.format(version.uid))
                continue

            if not version.status.is_removable():
                logger.info('Not considering version {}, it has a status of {}.'.format(
                    version.uid, version.status.name))
                continue

            versions_by_volume[version.volume].append(version)

        dismissed_versions: Set[Version] = set()
        for versions_slice in versions_by_volume.values():
            dismissed_versions |= set(RetentionFilter(rules_spec).filter(versions_slice))

        if dismissed_versions and group_label is not None:
            additional_versions: Set[Version] = set()
            for version in dismissed_versions:
                label_match = list(filter(lambda label: label.name == group_label, version.labels.values()))
                if not label_match:
                    continue
                assert len(label_match) == 1
                additional_versions |= set(Version.find(labels=[(group_label, label_match[0].value)]))
            dismissed_versions |= additional_versions

        if dismissed_versions:
            logger.info('Removing versions: {}.'.format(', '.join(
                map(lambda version: version.uid, sorted(dismissed_versions)))))
        else:
            logger.info('All versions are conforming to the retention policy.')

        if dry_run:
            logger.info('Dry run, will not remove anything.')
            return []

        # Iterate through copy of dismissed_versions
        for version in list(dismissed_versions):
            try:
                self.rm(version.uid, force=True, keep_metadata_backup=keep_metadata_backup)
            except KeyError:
                logger.warning(f'Version {version.uid} was removed in the meantime.')
                dismissed_versions.remove(version)
            except PermissionError as exception:
                logger.warning(str(exception))
                dismissed_versions.remove(version)
            except AlreadyLocked:
                logger.warning(f'Version {version.uid} could not be deleted, it is currently locked.')
                dismissed_versions.remove(version)

        return sorted(dismissed_versions)

    def storage_stats(self, storage_name: str = None) -> Tuple[int, int]:
        storage = StorageFactory.get_by_name(storage_name or self._default_storage_name)
        return storage.storage_stats()

    @staticmethod
    def list_storages() -> List[str]:
        return list(StorageFactory.get_modules().keys())

    @staticmethod
    def storage_usage(filter_expression: str = None) -> Dict[str, Dict[str, int]]:
        return Version.storage_usage(filter_expression)


class _BlockStore:

    _block_present: Set[BlockUid]

    def __init__(self, directory: str) -> None:
        self._directory = directory
        self._block_present = set()

    def _cache_filename(self, block_uid: BlockUid) -> str:
        assert block_uid.left is not None and block_uid.right is not None
        filename = '{:016x}-{:016x}'.format(block_uid.left, block_uid.right)
        digest = hashlib.md5(filename.encode('ascii')).hexdigest()
        return os.path.join(self._directory, '{}/{}/{}'.format(digest[0:2], digest[2:4], filename))

    def read(self, block_uid: BlockUid, offset: int = 0, length: int = None) -> bytes:
        filename = self._cache_filename(block_uid)
        with open(filename, 'rb', buffering=0) as f:
            f.seek(offset)
            if length is None:
                return f.read()
            else:
                return f.read(length)

    def write(self, block_uid: BlockUid, data) -> None:
        filename = self._cache_filename(block_uid)
        try:
            with open(filename, 'wb', buffering=0) as f:
                f.write(data)
        except FileNotFoundError:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'wb', buffering=0) as f:
                f.write(data)

        self._block_present.add(block_uid)

    def update(self, block_uid: BlockUid, offset: int, data: bytes) -> None:
        filename = self._cache_filename(block_uid)
        with open(filename, 'r+b', buffering=0) as f:
            f.seek(offset)
            f.write(data)

    def rm(self, block_uid: BlockUid) -> None:
        try:
            self._block_present.remove(block_uid)
        except KeyError:
            pass
        filename = self._cache_filename(block_uid)
        try:
            os.unlink(filename)
        except FileNotFoundError:
            pass

    def present(self, block_uid: BlockUid) -> bool:
        return block_uid in self._block_present


# The reason for this class being here is that it accesses private attributes of class Benji
# and I don't want to make them all generally publicly available.
# Maybe they could inherit from the same base class in the future, but currently their
# functionality seems very different. So we just define that BenjiStore objects may access
# private attributes of Benji objects.
class BenjiStore(ReprMixIn):

    _benji_obj: Benji
    _cow: Dict[VersionUid, Dict[int, DereferencedBlock]]

    def __init__(self, benji_obj: Benji) -> None:
        self._benji_obj = benji_obj
        self._cow = {}  # contains version_uid: dict() of block id -> block

        block_cache_directory = self._benji_obj.config.get('nbd.blockCache.directory', types=str)
        block_cache_maximum_size = self._benji_obj.config.get('nbd.blockCache.maximumSize', types=int)
        cow_store_directory = self._benji_obj.config.get('nbd.cowStore.directory', types=str)

        os.makedirs(block_cache_directory, exist_ok=True)
        self._block_cache = Cache(block_cache_directory,
                                  size_limit=block_cache_maximum_size,
                                  eviction_policy='least-frequently-used',
                                  disk_min_file_size=0)

        os.makedirs(cow_store_directory, exist_ok=True)
        self._cow_store = _BlockStore(cow_store_directory)

    @staticmethod
    def open(version) -> None:
        Locking.lock_version(version.uid, reason='NBD')

    @staticmethod
    def close(version) -> None:
        Locking.unlock_version(version.uid)

    @staticmethod
    def find_versions(version_uid: VersionUid = None) -> List[Version]:
        return Version.find(version_uid=version_uid)

    @staticmethod
    def _block_list(version: Version, offset: int, length: int) -> List[Tuple[Optional[Block], int, int]]:
        block_idx = offset // version.block_size
        block_offset = offset % version.block_size

        chunks: List[Tuple[Optional[Block], int, int]] = []
        while True:
            block = version.get_block_by_idx(block_idx)
            if block is None:
                # We round up the size reported by the NBD server to a multiple of 4096 which is the maximum
                # block size supported by NBD. So we might need to fake up to 4095 bytes (of zeros) here.
                if length > 4095:
                    # Don't throw one of our own exceptions here as we need an exception with an errno value
                    # to communicate it back in the NBD response.
                    raise OSError(errno.EIO)
                length_in_block = min(block.size - block_offset, length)
                chunks.append((None, 0, length_in_block))
            else:
                length_in_block = min(block.size - block_offset, length)
                chunks.append((block, block_offset, length_in_block))

            block_idx += 1
            block_offset = 0
            length -= length_in_block

            assert length >= 0
            if length == 0:
                break

        return chunks

    def read(self, version: Version, cow_version: Optional[Version], offset: int, length: int) -> bytes:
        if cow_version:
            cow: Optional[Dict[int, DereferencedBlock]] = self._cow[cow_version.uid]
        else:
            cow = None
        read_list = self._block_list(version, offset, length)
        data_chunks: List[bytes] = []
        block: Optional[Union[Block, DereferencedBlock]]
        for block, offset_in_block, length_in_block in read_list:
            # Access lies beyond end of version
            if block is None:
                logger.warning('Tried to read data beyond device (version {}, size {}, offset {}).'.format(
                    version.uid, version.size, offset_in_block))
                data_chunks.append(b'\0' * length_in_block)
                continue

            if cow is not None and block.idx in cow:
                # Read block from COW
                assert cow_version is not None

                block = cow[block.idx]
                logger.debug('Reading block from COW {}/{} {}:{}.'.format(cow_version.uid, block.idx, offset_in_block,
                                                                          length_in_block))

                assert self._cow_store.present(block.uid)
                data_chunks.append(self._cow_store.read(block.uid, offset_in_block, length_in_block))
            else:
                # Read block from original version (if not sparse)
                logger.debug('Reading {}block {}/{} {}:{}.'.format('sparse ' if not block.uid else '', version.uid,
                                                                   block.idx, offset_in_block, length_in_block))

                # Block is sparse
                if not block.uid:
                    data_chunks.append(b'\0' * length_in_block)
                    continue

                block_f = self._block_cache.get(str(block.uid), read=True)
                if block_f is not None:
                    # Cache hit
                    try:
                        block_f.seek(offset_in_block)
                        data_chunks.append(block_f.read(length_in_block))
                    finally:
                        block_f.close()
                else:
                    storage = StorageFactory.get_by_name(version.storage.name)
                    data = storage.read_block(block)
                    self._block_cache[str(block.uid)] = data
                    data_chunks.append(data[offset_in_block:offset_in_block + length_in_block])

        return b''.join(data_chunks)

    def create_cow_version(self, base_version: Version) -> Version:
        cow_version = self._benji_obj.create_version(
            version_uid=VersionUid('{}-{}'.format(f'nbd-cow-{base_version.uid}'[:248], random_string(6))),
            volume=base_version.volume,
            snapshot=datetime.datetime.utcnow().isoformat(timespec='microseconds') + 'Z',
            base_version_uid=base_version.uid,
            base_version_locking=False)
        Locking.update_version_lock(cow_version.uid, reason='NBD COW')
        self._cow[cow_version.uid] = {}  # contains version_uid: dict() of block id -> uid
        return cow_version

    def write(self, cow_version: Version, offset: int, data: bytes) -> None:
        """ Copy on write backup writer """
        cow = self._cow[cow_version.uid]
        write_list = self._block_list(cow_version, offset, len(data))
        position_in_data = 0
        for block, offset_in_block, length_in_block in write_list:
            if block is None:
                logger.warning('Tried to write data beyond device, it will be lost (version {}, size {}, offset {}).'.format(
                    cow_version.uid, cow_version.size, offset))
                break
            if block.idx in cow:
                # The block is already copied, so update in the cache
                update_block = cow[block.idx]
                self._cow_store.update(update_block.uid, offset_in_block,
                                       data[position_in_data:position_in_data + length_in_block])
                logger.debug('Updated block {}/{} {}:{}.'.format(cow_version.uid, block.idx, offset_in_block,
                                                                 length_in_block))
            else:
                # Read the block from the original
                if block.uid:
                    storage = StorageFactory.get_by_name(cow_version.storage.name)
                    write_data = BytesIO(storage.read_block(block))
                # Was a sparse block
                else:
                    write_data = BytesIO(b'\0' * cow_version.block_size)

                # Update the block
                write_data.seek(offset_in_block)
                write_data.write(data[position_in_data:position_in_data + length_in_block])
                write_data.seek(0)

                # Save a copy of the changed data and record the changed block UID
                new_block = block.deref()
                new_block.uid = BlockUid(cow_version.id, block.idx + 1)
                new_block.checksum = None
                self._cow_store.write(new_block.uid, write_data.read())
                cow[block.idx] = new_block
                logger.debug('COW: Wrote block {}/{} {}:{} into {}.'.format(cow_version.uid, block.idx, offset_in_block,
                                                                            length_in_block, new_block.uid))
            position_in_data += length_in_block

    def flush(self, cow_version: Version) -> None:
        pass

    def fixate(self, cow_version: Version) -> None:
        # save blocks into version
        logger.info('Fixating version {} with {} blocks, please wait.'.format(cow_version.uid,
                                                                              len(self._cow[cow_version.uid])))

        sparse_block_checksum = self._benji_obj._block_hash.data_hexdigest(b'\0' * cow_version.block_size)
        storage = StorageFactory.get_by_name(cow_version.storage.name)
        for block in self._cow[cow_version.uid].values():
            logger.debug('Fixating block {}/{} with UID {}'.format(cow_version.uid, block.idx, block.uid))
            data = self._cow_store.read(block.uid)

            block.checksum = self._benji_obj._block_hash.data_hexdigest(data)
            if block.checksum == sparse_block_checksum:
                logger.debug('Detected sparse block {}/{}.'.format(cow_version.uid, block.idx))
                # The remove assumes that each block UID appears only once in the list and is not shared in any way.
                self._cow_store.rm(block.uid)
                block.checksum = None
                block.uid = BlockUid(None, None)
            else:
                storage.write_block(block, data)
                # The remove assumes that each block UID appears only once in the list and is not shared in any way.
                self._cow_store.rm(block.uid)

            try:
                cow_version.set_block(idx=block.idx,
                                      block_uid=block.uid,
                                      checksum=block.checksum,
                                      size=len(data),
                                      valid=True)
            except:
                # Prevent orphaned blocks
                if block.uid:
                    storage.rm_block(block.uid)

        cow_version.commit()
        cow_version.set(status=VersionStatus.valid, protected=True)
        self._benji_obj.metadata_backup([cow_version.uid], overwrite=True, locking=False)
        Locking.unlock_version(cow_version.uid)
        del self._cow[cow_version.uid]
        logger.info('Finished.')



import importlib
from typing import NamedTuple, Any, Dict

from benji.config import Config, ConfigList
from benji.exception import InternalError, ConfigurationError
from benji.repr import ReprMixIn


class _TransformFactoryModule(NamedTuple):
    module: Any
    arguments: Dict[str, Any]


class TransformFactory(ReprMixIn):

    _modules: Dict[str, _TransformFactoryModule] = {}
    _instances: Dict[str, Any] = {}

    def __init__(self) -> None:
        raise InternalError('TransformFactory constructor called.')

    @classmethod
    def _import_modules(cls, config: Config, modules: ConfigList) -> None:
        for index, module_dict in enumerate(modules):
            module = Config.get_from_dict(module_dict,
                                          'module',
                                          types=str,
                                          full_name_override=modules.full_name,
                                          index=index)
            name = Config.get_from_dict(module_dict,
                                        'name',
                                        types=str,
                                        full_name_override=modules.full_name,
                                        index=index)
            configuration = Config.get_from_dict(module_dict,
                                                 'configuration',
                                                 None,
                                                 types=dict,
                                                 full_name_override=modules.full_name,
                                                 index=index)

            if name in cls._modules:
                raise ConfigurationError('Duplicate name "{}" in list {}.'.format(name, modules.full_name))

            module = importlib.import_module('{}.{}'.format(__package__, module))
            try:
                configuration = config.validate(module=module.__name__, config=configuration)
            except ConfigurationError as exception:
                raise ConfigurationError('Configuration for transform {} is invalid.'.format(name)) from exception
            cls._modules[name] = _TransformFactoryModule(module=module,
                                                         arguments={
                                                             'config': config,
                                                             'name': name,
                                                             'module_configuration': configuration
                                                         })

    @classmethod
    def initialize(cls, config: Config) -> None:
        cls._modules = {}
        transforms: ConfigList = config.get('transforms', None, types=list)
        if transforms is not None:
            cls._import_modules(config, transforms)

    @classmethod
    def close(cls) -> None:
        cls._instances = {}

    @classmethod
    def get_by_name(cls, name: str) -> Any:
        if name not in cls._instances:
            if name not in cls._modules:
                raise ConfigurationError('Transform name {} is undefined.'.format(name))

            module = cls._modules[name].module
            module_arguments = cls._modules[name].arguments
            cls._instances[name] = module.Transform(**module_arguments)

        return cls._instances[name]



import datetime
import math
import time
import timeit
import uuid
from typing import List, Dict, Any
from unittest import TestCase

from dateutil import tz

from benji.database import BlockUid, VersionUid, VersionStatus, Version, Storage, DeletedBlock, Locking
from benji.exception import InternalError, UsageError, AlreadyLocked
from benji.logging import logger
from benji.tests.testcase import DatabaseBackendTestCaseBase


class DatabaseBackendTestCase(DatabaseBackendTestCaseBase):

    def test_version(self):
        Storage.sync('s-1', storage_id=1)
        version = Version.create(version_uid=VersionUid('v1'),
                                 volume='backup-name',
                                 snapshot='snapshot-name',
                                 size=16 * 1024 * 4096,
                                 storage_id=1,
                                 block_size=4 * 1024 * 4096)

        version = Version.get_by_uid(version.uid)
        self.assertEqual('backup-name', version.volume)
        self.assertEqual('snapshot-name', version.snapshot)
        self.assertEqual(16 * 1024 * 4096, version.size)
        self.assertEqual(4 * 1024 * 4096, version.block_size)
        self.assertEqual(version.status, VersionStatus.incomplete)
        self.assertFalse(version.protected)

        version.set(status=VersionStatus.valid)
        version = Version.get_by_uid(version.uid)
        self.assertEqual(version.status, VersionStatus.valid)

        version.set(status=VersionStatus.invalid)
        version = Version.get_by_uid(version.uid)
        self.assertEqual(version.status, VersionStatus.invalid)

        version.set(protected=True)
        version = Version.get_by_uid(version.uid)
        self.assertTrue(version.protected)

        version.set(protected=False)
        version = Version.get_by_uid(version.uid)
        self.assertFalse(version.protected)

        version.add_label('label-1', 'bla')
        version.add_label('label-2', '')
        version = Version.get_by_uid(version.uid)
        self.assertEqual(2, len(version.labels))
        self.assertEqual(version.id, version.labels['label-1'].version_id)
        self.assertEqual('label-1', version.labels['label-1'].name)
        self.assertEqual('bla', version.labels['label-1'].value)
        self.assertEqual(version.id, version.labels['label-2'].version_id)
        self.assertEqual('label-2', version.labels['label-2'].name)
        self.assertEqual('', version.labels['label-2'].value)

        version.add_label('label-2', 'test123')
        version = Version.get_by_uid(version.uid)
        self.assertEqual(version.id, version.labels['label-2'].version_id)
        self.assertEqual('label-2', version.labels['label-2'].name)
        self.assertEqual('test123', version.labels['label-2'].value)

        version.rm_label('label-1')
        version = Version.get_by_uid(version.uid)
        self.assertEqual(1, len(version.labels))

        version.rm_label('label-2')
        version = Version.get_by_uid(version.uid)
        self.assertEqual(0, len(version.labels))

        version.rm_label('label-3')
        version = Version.get_by_uid(version.uid)
        self.assertEqual(0, len(version.labels))

    def test_block(self):
        Storage.sync('s-1', storage_id=1)
        version = Version.create(version_uid=VersionUid('v1'),
                                 volume='name-' + self.random_string(12),
                                 snapshot='snapshot-name-' + self.random_string(12),
                                 size=256 * 1024 * 4096,
                                 block_size=1024 * 4096,
                                 storage_id=1)

        checksums = []
        uids = []
        num_blocks = 256
        blocks: List[Dict[str, Any]] = []
        for idx in range(num_blocks):
            checksums.append(self.random_hex(64))
            uids.append(BlockUid(1, idx))
            blocks.append({
                'idx': idx,
                'uid_left': uids[idx].left,
                'uid_right': uids[idx].right,
                'checksum': checksums[idx],
                'size': 1024 * 4096,
                'valid': True
            })
        version.create_blocks(blocks=blocks)

        for idx, checksum in enumerate(checksums):
            block = version.get_block_by_checksum(checksum)
            self.assertEqual(idx, block.idx)
            self.assertEqual(version.id, block.version_id)
            self.assertEqual(uids[idx], block.uid)
            self.assertEqual(checksum, block.checksum)
            self.assertEqual(1024 * 4096, block.size)
            self.assertTrue(block.valid)

        for idx, uid in enumerate(uids):
            block = version.get_block_by_idx(idx)
            self.assertEqual(idx, block.idx)
            self.assertEqual(version.id, block.version_id)
            self.assertEqual(uid, block.uid)
            self.assertEqual(checksums[idx], block.checksum)
            self.assertEqual(1024 * 4096, block.size)
            self.assertTrue(block.valid)

        self.assertEqual(num_blocks, len(list(version.blocks)))
        self.assertEqual(num_blocks, version.blocks_count)
        self.assertEqual(0, version.sparse_blocks_count)

        for idx, block in enumerate(version.blocks):
            self.assertEqual(idx, block.idx)
            self.assertEqual(version.id, block.version_id)
            self.assertEqual(uids[idx], block.uid)
            self.assertEqual(checksums[idx], block.checksum)
            self.assertEqual(1024 * 4096, block.size)
            self.assertTrue(block.valid)

        for idx, block in enumerate(version.blocks):
            dereferenced_block = block.deref()
            self.assertEqual(idx, dereferenced_block.idx)
            self.assertEqual(version.id, dereferenced_block.version_id)
            self.assertEqual(uids[idx].left, dereferenced_block.uid.left)
            self.assertEqual(uids[idx].right, dereferenced_block.uid.right)
            self.assertEqual(checksums[idx], dereferenced_block.checksum)
            self.assertEqual(1024 * 4096, dereferenced_block.size)
            self.assertTrue(dereferenced_block.valid)

        version.remove()
        deleted_count = 0
        for uids_deleted in DeletedBlock.get_unused_block_uids(-1):
            for storage in uids_deleted.values():
                for uid in storage:
                    self.assertIn(uid, uids)
                    deleted_count += 1
        self.assertEqual(num_blocks, deleted_count)

    def test_lock_version(self):
        Locking.lock_version(VersionUid('v1'), reason='locking test')
        self.assertRaises(InternalError, lambda: Locking.lock_version(VersionUid('v1'), reason='locking test'))
        Locking.unlock_version(VersionUid('v1'))

    def test_is_locked(self):
        Locking.lock(lock_name='test', reason='locking test')
        self.assertTrue(Locking.is_locked(lock_name='test'))
        Locking.unlock(lock_name='test')
        self.assertFalse(Locking.is_locked(lock_name='test'))

    def test_is_version_locked(self):
        Locking.lock_version(VersionUid('v1'), reason='locking test')
        self.assertTrue(Locking.is_version_locked(VersionUid('v1')))
        Locking.unlock_version(VersionUid('v1'))
        self.assertFalse(Locking.is_version_locked(VersionUid('v1')))

    def test_lock_version_context_manager(self):
        with Locking.with_version_lock(VersionUid('v1'), reason='locking test'):
            with self.assertRaises(InternalError):
                Locking.lock_version(VersionUid('v1'), reason='locking test')
        Locking.lock_version(VersionUid('v1'), reason='locking test')
        Locking.unlock_version(VersionUid('v1'))

    def test_lock_context_manager(self):
        with Locking.with_lock(lock_name='test', reason='locking test'):
            with self.assertRaises(InternalError):
                Locking.lock(lock_name='test', reason='locking test')
        Locking.lock(lock_name='test', reason='locking test')
        Locking.unlock(lock_name='test')

    def test_lock_override(self):
        Locking.lock_version(VersionUid('v1'), reason='locking test')
        self.assertRaises(InternalError, lambda: Locking.lock_version(VersionUid('v1'), reason='locking test'))
        old_uuid = Locking._uuid
        new_uuid = uuid.uuid1().hex
        # This fakes the appearance of another instance
        Locking._uuid = new_uuid
        self.assertRaises(AlreadyLocked, lambda: Locking.lock_version(VersionUid('v1'), reason='locking test'))
        Locking.lock_version(VersionUid('v1'), reason='locking test', override_lock=True)
        self.assertRaises(InternalError, lambda: Locking.lock_version(VersionUid('v1'), reason='locking test'))
        Locking._uuid = old_uuid
        self.assertRaises(AlreadyLocked, lambda: Locking.lock_version(VersionUid('v1'), reason='locking test'))
        Locking._uuid = new_uuid
        Locking.unlock_version(VersionUid('v1'))

    def test_version_filter(self):
        Storage.sync('s-1', storage_id=1)
        for i in range(256):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=16 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096,
                                     status=VersionStatus.valid)
            version = Version.get_by_uid(version.uid)
            self.assertEqual(0, len(version.labels))
            version.add_label('label-key', 'label-value')
            version.add_label('label-key-2', str(i))
            version.add_label('label-key-3', '')
            if i > 127:
                version.add_label('label-key-4', '')
                self.assertEqual(4, len(version.labels))
            else:
                self.assertEqual(3, len(version.labels))

        versions = Version.find_with_filter()
        self.assertEqual(256, len(versions))

        versions = Version.find_with_filter('labels["label-key"] == "label-value"')
        self.assertEqual(256, len(versions))

        versions = Version.find_with_filter('"label-value" == labels["label-key"]')
        self.assertEqual(256, len(versions))

        self.assertRaises(UsageError, lambda: Version.find_with_filter('labels["label-key"] and "label-value"'))
        self.assertRaises(UsageError, lambda: Version.find_with_filter('True'))
        self.assertRaises(UsageError, lambda: Version.find_with_filter('10'))
        self.assertRaises(UsageError, lambda: Version.find_with_filter('labels["label-key"] == "label-value" and True'))
        self.assertRaises(UsageError, lambda: Version.find_with_filter('labels["label-key"] == "label-value" and False'))
        self.assertRaises(UsageError, lambda: Version.find_with_filter('"hallo" == "hey"'))

        # volume is always true because it is never empty
        versions = Version.find_with_filter('volume')
        self.assertEqual(256, len(versions))

        versions = Version.find_with_filter('status == "valid"')
        self.assertEqual(256, len(versions))

        self.assertRaises(UsageError, lambda: Version.find_with_filter('status == wrong'))

        versions = Version.find_with_filter('labels["label-key-3"] == ""')
        self.assertEqual(256, len(versions))

        versions = Version.find_with_filter('labels["label-key"] != "label-value"')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('labels["label-key-2"] == 9')
        self.assertEqual(1, len(versions))

        versions = Version.find_with_filter('snapshot == "snapshot-name.1"')
        self.assertEqual(1, len(versions))
        self.assertEqual(VersionUid('v2'), versions[0].uid)

        versions = Version.find_with_filter('snapshot == "snapshot-name.1" and labels["label-key-2"] == 1')
        self.assertEqual(1, len(versions))
        self.assertEqual(VersionUid('v2'), versions[0].uid)

        versions = Version.find_with_filter('snapshot == "snapshot-name.1" and labels["label-key-2"] == "2"')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('snapshot == "snapshot-name.1" or labels["label-key-2"] == 2')
        self.assertEqual(2, len(versions))
        self.assertSetEqual({VersionUid('v2'), VersionUid('v3')}, {version.uid for version in versions})

        versions = Version.find_with_filter('volume == "backup-name" and snapshot == "snapshot-name.1"')
        self.assertEqual(1, len(versions))
        self.assertEqual(VersionUid('v2'), versions[0].uid)

        versions = Version.find_with_filter('volume == "backup-name" and (snapshot == "snapshot-name.1" or snapshot == "snapshot-name.2")')
        self.assertEqual(2, len(versions))
        self.assertSetEqual({VersionUid('v2'), VersionUid('v3')}, {version.uid for version in versions})

        versions = Version.find_with_filter('uid == "v1" or uid == "v12"')
        self.assertEqual(2, len(versions))
        self.assertSetEqual({VersionUid('v1'), VersionUid('v12')}, {version.uid for version in versions})

        versions = Version.find_with_filter('uid == "v1" and uid == "v12"')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('not labels["not-exists"]')
        self.assertEqual(256, len(versions))

        versions = Version.find_with_filter('labels["label-key-4"]')
        self.assertEqual(128, len(versions))

        versions = Version.find_with_filter('labels["label-key-4"] and volume')
        self.assertEqual(128, len(versions))

    # Issue https://github.com/elemental-lf/benji/issues/9
    def test_version_filter_issue_9(self):
        Storage.sync('s-1', storage_id=1)
        version_uids = set()
        for i in range(3):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=16 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096,
                                     status=VersionStatus.valid)
            self.assertNotIn(version.uid, version_uids)
            version_uids.add(version.uid)

        versions = Version.find_with_filter('snapshot == "snapshot-name.2" and volume == "backup-name" and status == "valid"')
        self.assertEqual(1, len(versions))

        versions = Version.find_with_filter('snapshot == "snapshot-name.2" or volume == "backup-name" or status == "valid"')
        self.assertEqual(3, len(versions))

    # Issue https://github.com/elemental-lf/benji/issues/9 (slowness part)
    def test_version_filter_issue_9_slowness(self):
        Storage.sync('s-1', storage_id=1)
        version_uids = set()
        for i in range(3):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=16 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096)
            self.assertNotIn(version.uid, version_uids)
            version_uids.add(version.uid)

        t1 = timeit.timeit(lambda: Version.find_with_filter('snapshot == "snapshot-name.2" and volume == "backup-name"'),
                           number=1)
        t2 = timeit.timeit(lambda: Version.find_with_filter('(snapshot == "snapshot-name.2" and volume == "backup-name")'),
                           number=1)
        logger.debug('test_version_filter_issue_9_slowness: t1 {}, t2 {}'.format(t1, t2))
        self.assertLess(t1 - t2, 5)

    def test_version_filter_dateparse(self):
        Storage.sync('s-1', storage_id=1)
        version_uids = set()
        for i in range(3):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=16 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096)
            self.assertNotIn(version.uid, version_uids)
            version_uids.add(version.uid)

        # Wait at least one seconds
        time.sleep(1)

        versions = Version.find_with_filter('date <= "now"')
        self.assertEqual(3, len(versions))

        versions = Version.find_with_filter('date > "1 month ago"')
        self.assertEqual(3, len(versions))

        versions = Version.find_with_filter('date > "now"')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('date < "1 month ago"')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('"now" >= date')
        self.assertEqual(3, len(versions))

        versions = Version.find_with_filter('"1 month ago" < date')
        self.assertEqual(3, len(versions))

        versions = Version.find_with_filter('"now" < date')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('"1 month ago" > date')
        self.assertEqual(0, len(versions))

        versions = Version.find_with_filter('date <= "{}"'.format(
            datetime.datetime.now(tz=tz.tzlocal()).strftime("%Y-%m-%dT%H:%M:%S")))
        self.assertEqual(3, len(versions))

    def test_set_block_valid(self):
        Storage.sync('s-1', storage_id=1)
        versions = []
        good_uid = BlockUid(1, 2)
        bad_uid = BlockUid(3, 4)
        for i in range(6):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=16 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096)
            blocks = [{
                'idx': 0,
                'uid_left': bad_uid.left if i < 3 else good_uid.left,
                'uid_right': bad_uid.right if i < 3 else good_uid.right,
                'checksum': 'aabbcc',
                'size': 4 * 1024 * 4096,
                'valid': True,
            }]
            version.create_blocks(blocks=blocks)
            version.set(status=VersionStatus.valid)

            versions.append(version)

        for i in range(6):
            self.assertEqual(VersionStatus.valid, versions[i].status)
            self.assertTrue(list(versions[i].blocks)[0].valid)

        Version.set_block_valid(bad_uid, False)

        for i in range(3):
            self.assertEqual(VersionStatus.invalid, versions[i].status)
            self.assertFalse(list(versions[i].blocks)[0].valid)

        for i in range(3, 6):
            self.assertEqual(VersionStatus.valid, versions[i].status)
            self.assertTrue(list(versions[i].blocks)[0].valid)

        Version.set_block_valid(bad_uid, True)

        for i in range(3):
            self.assertEqual(VersionStatus.invalid, versions[i].status)
            self.assertTrue(list(versions[i].blocks)[0].valid)

        for i in range(3, 6):
            self.assertEqual(VersionStatus.valid, versions[i].status)
            self.assertTrue(list(versions[i].blocks)[0].valid)

    def test_version_blocks_count(self):
        Storage.sync('s-1', storage_id=1)
        for i in range(256):
            version = Version.create(version_uid=f'v{i + 1}',
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=i * 1111 * 4096,
                                     storage_id=1,
                                     block_size=4 * 1024 * 4096)
            self.assertEqual(math.ceil(version.size / version.block_size), version.blocks_count)

    def test_storage_usage(self):
        Storage.sync('s-1', storage_id=1)
        for i in range(2):
            version = Version.create(version_uid=VersionUid(f'v{i + 1}'),
                                     volume='backup-name',
                                     snapshot='snapshot-name.{}'.format(i),
                                     size=32 * 1024 * 4096,
                                     storage_id=1,
                                     block_size=1024 * 4096)

            blocks: List[Dict[str, Any]] = []
            # shared
            for idx in range(0, 6):
                uid = BlockUid(1, idx)
                blocks.append({
                    'idx': idx,
                    'uid_left': uid.left,
                    'uid_right': uid.right,
                    'checksum': None,
                    'size': 1024 * 4096,
                    'valid': True
                })
            # sparse
            for idx in range(6, 13):
                uid = BlockUid(i + 1, idx)
                blocks.append({
                    'idx': idx,
                    'uid_left': None,
                    'uid_right': None,
                    'checksum': None,
                    'size': 1024 * 4096,
                    'valid': True
                })
            # exclusive
            for idx in range(13, 25):
                uid = BlockUid(i + 1, idx)
                blocks.append({
                    'idx': idx,
                    'uid_left': uid.left,
                    'uid_right': uid.right,
                    'checksum': None,
                    'size': 1024 * 4096,
                    'valid': True
                })
            # exclusive deduplicated
            for idx in range(25, 32):
                uid = BlockUid(i + 1, idx - 7)
                blocks.append({
                    'idx': idx,
                    'uid_left': uid.left,
                    'uid_right': uid.right,
                    'checksum': None,
                    'size': 1024 * 4096,
                    'valid': True
                })
            version.create_blocks(blocks=blocks)

        for uid in ('v1', 'v2'):
            usage = Version.storage_usage(f'uid == "{uid}"')
            self.assertIsInstance(usage.get('s-1', None), dict)
            usage_s_1 = usage['s-1']
            check_value_matrix = (
                ('virtual', 32 * 1024 * 4096),
                ('shared', 6 * 1024 * 4096),
                ('sparse', 7 * 1024 * 4096),
                ('exclusive', 19 * 1024 * 4096),
                ('deduplicated_exclusive', (19 - 7) * 1024 * 4096),
            )
            for field, amount in check_value_matrix:
                value = usage_s_1.get(field, None)
                self.assertIsInstance(value, int)
                self.assertEqual(amount, value)


class DatabaseBackendTestSQLLite(DatabaseBackendTestCase, TestCase):

    CONFIG = """
        configurationVersion: '1'
        logFile: /dev/stderr
        ios:
        - name: file
          module: file
        defaultStorage: s1
        storages:
        - name: s1
          storageId: 1
          module: file
          configuration:
            path: {testpath}/data
        databaseEngine: sqlite:///{testpath}/benji.sqlite
        """


class DatabaseBackendTestSQLLiteInMemory(DatabaseBackendTestCase, TestCase):

    CONFIG = """
        configurationVersion: '1'
        logFile: /dev/stderr
        ios:
        - name: file
          module: file
        defaultStorage: s1
        storages:
        - name: s1
          storageId: 1
          module: file
          configuration:
            path: {testpath}/data
        databaseEngine: sqlite://       
        """


class DatabaseBackendTestPostgreSQL(DatabaseBackendTestCase, TestCase):

    CONFIG = """
        configurationVersion: '1'
        logFile: /dev/stderr
        ios:
        - name: file
          module: file
        defaultStorage: s1
        storages:
        - name: s1
          storageId: 1
          module: file
          configuration:
            path: {testpath}/data
        databaseEngine: postgresql://benji:verysecret@localhost:15432/benji
        """



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import queue
import re
import threading
import time
from collections import deque
from typing import Tuple, Optional, Union, Iterator, Deque
from urllib.parse import parse_qs

import rados

# noinspection PyUnresolvedReferences
import rbd
from benji.config import ConfigDict, Config
from benji.database import DereferencedBlock, Block
from benji.exception import UsageError, ConfigurationError
from benji.io.base import IOBase
from benji.logging import logger


class IO(IOBase):

    _pool_name: Optional[str]
    _image_name: Optional[str]
    _snapshot_name: Optional[str]

    def __init__(self, *, config: Config, name: str, module_configuration: ConfigDict, url: str,
                 block_size: int) -> None:
        super().__init__(config=config,
                         name=name,
                         module_configuration=module_configuration,
                         url=url,
                         block_size=block_size)

        if self.parsed_url.username or self.parsed_url.password or self.parsed_url.hostname or self.parsed_url.port \
                    or self.parsed_url.params or self.parsed_url.fragment:
            raise UsageError('The supplied URL {} is invalid.'.format(self.url))
        if self.parsed_url.query:
            try:
                extra_ceph_conf = parse_qs(self.parsed_url.query,
                                           keep_blank_values=True,
                                           strict_parsing=True,
                                           errors='strict')
            except (ValueError, UnicodeError) as exception:
                raise UsageError('The supplied URL {} is invalid.'.format(self.url)) from exception

            # parse_qs returns the values as lists, only consider the first appearance of each key in the query string.
            extra_ceph_conf = {key: value[0] for key, value in extra_ceph_conf.items()}
        else:
            extra_ceph_conf = {}

        ceph_config_file = config.get_from_dict(module_configuration, 'cephConfigFile', types=str)
        if 'client_identifier' in extra_ceph_conf:
            client_identifier = extra_ceph_conf['client_identifier']
            del extra_ceph_conf['client_identifier']
        else:
            client_identifier = config.get_from_dict(module_configuration, 'clientIdentifier', types=str)

        self._cluster = rados.Rados(conffile=ceph_config_file, rados_id=client_identifier, conf=extra_ceph_conf)
        self._cluster.connect()
        # create a bitwise or'd list of the configured features
        self._new_image_features = 0
        for feature in config.get_from_dict(module_configuration, 'newImageFeatures', types=list):
            try:
                self._new_image_features = self._new_image_features | getattr(rbd, feature)
            except AttributeError:
                raise ConfigurationError('{}: Unknown image feature {}.'.format(module_configuration.full_name, feature))

        self._pool_name = None
        self._image_name = None
        self._snapshot_name = None
        self._rbd_image = None

        self._simultaneous_reads = config.get_from_dict(module_configuration, 'simultaneousReads', types=int)
        self._simultaneous_writes = config.get_from_dict(module_configuration, 'simultaneousWrites', types=int)
        self._read_queue: Deque[DereferencedBlock] = deque()
        self._write_queue: Deque[Tuple[DereferencedBlock, bytes]] = deque()
        self._outstanding_aio_reads = 0
        self._outstanding_aio_writes = 0
        self._submitted_aio_writes = threading.BoundedSemaphore(self._simultaneous_writes)
        self._read_completion_queue: queue.Queue[Tuple[rbd.Completion, float, float, DereferencedBlock,
                                                       bytes]] = queue.Queue()
        self._write_completion_queue: queue.Queue[Tuple[rbd.Completion, float, float,
                                                        DereferencedBlock]] = queue.Queue()

    def open_r(self) -> None:
        re_match = re.match('^([^/]+)/([^@]+)(?:@(.+))?$', self.parsed_url.path)
        if not re_match:
            raise UsageError('URL {} is invalid . Need {}:<pool>/<imagename> or {}:<pool>/<imagename>@<snapshotname>.'.format(
                self.url, self.name, self.name))
        self._pool_name, self._image_name, self._snapshot_name = re_match.groups()

        # try opening it and quit if that's not possible.
        try:
            ioctx = self._cluster.open_ioctx(self._pool_name)
        except rados.ObjectNotFound:
            raise FileNotFoundError('Ceph pool {} not found.'.format(self._pool_name)) from None

        try:
            self._rbd_image = rbd.Image(ioctx, self._image_name, self._snapshot_name, read_only=True)
        except rbd.ImageNotFound:
            raise FileNotFoundError('RBD image or snapshot {} not found.'.format(self.url)) from None

    def open_w(self, size: int, force: bool = False, sparse: bool = False) -> None:
        re_match = re.match('^([^/]+)/([^@]+)$', self.parsed_url.path)
        if not re_match:
            raise UsageError('URL {} is invalid . Need {}:<pool>/<imagename>.'.format(self.url, self.name))
        self._pool_name, self._image_name = re_match.groups()

        # try opening it and quit if that's not possible.
        try:
            ioctx = self._cluster.open_ioctx(self._pool_name)
        except rados.ObjectNotFound:
            raise FileNotFoundError('Ceph pool {} not found.'.format(self._pool_name)) from None

        try:
            self._rbd_image = rbd.Image(ioctx, self._image_name)
        except rbd.ImageNotFound:
            rbd.RBD().create(ioctx, self._image_name, size, old_format=False, features=self._new_image_features)
            self._rbd_image = rbd.Image(ioctx, self._image_name)
        else:
            assert self._rbd_image is not None
            if not force:
                raise FileExistsError(
                    'RBD image {} already exists. Force the restore if you want to overwrite it.'.format(self.url))
            else:
                image_size = self._rbd_image.size()
                if size > image_size:
                    raise IOError(
                        'RBD image {} is too small. Its size is {} bytes, but we need {} bytes for the restore.'.format(
                            self.url, image_size, size))

                # If this is an existing image and sparse is true discard all objects from this image
                # RBD discard only supports a maximum region length of 0x7fffffff.
                if sparse:
                    logger.debug('Discarding all objects of RBD image {}.'.format(self.url))
                    region_start = 0
                    bytes_to_end = image_size
                    while bytes_to_end > 0:
                        region_length = min(0x7fffffff, bytes_to_end)
                        self._rbd_image.discard(region_start, region_length)
                        region_start += region_length
                        bytes_to_end -= region_length

    def close(self) -> None:
        assert self._rbd_image is not None
        self._rbd_image.close()

    def size(self) -> int:
        assert self._rbd_image is not None
        return self._rbd_image.size()

    def _submit_aio_reads(self):
        assert self._rbd_image is not None
        while len(self._read_queue) > 0 and self._outstanding_aio_reads < self._simultaneous_reads:
            block = self._read_queue.pop()
            t1 = time.time()

            def aio_callback(completion, data):
                t2 = time.time()
                self._read_completion_queue.put((completion, t1, t2, block, data))

            offset = block.idx * self.block_size
            self._rbd_image.aio_read(offset, block.size, aio_callback, rados.LIBRADOS_OP_FLAG_FADVISE_DONTNEED)
            self._outstanding_aio_reads += 1

    def read(self, block: Union[DereferencedBlock, Block]) -> None:
        self._read_queue.appendleft(block.deref())
        self._submit_aio_reads()

    def read_sync(self, block: Union[DereferencedBlock, Block]) -> bytes:
        assert self._rbd_image is not None
        offset = block.idx * self.block_size
        t1 = time.time()
        data = self._rbd_image.read(offset, block.size, rados.LIBRADOS_OP_FLAG_FADVISE_DONTNEED)
        t2 = time.time()

        if not data:
            raise EOFError('End of file reached on {} when there should be data.'.format(self.url))

        logger.debug('Read block {} in {:.3f}s'.format(block.idx, t2 - t1))

        return data

    def _reads_finished(self) -> bool:
        return len(self._read_queue) == 0 and self._outstanding_aio_reads == 0

    def read_get_completed(self,
                           timeout: Optional[int] = None
                          ) -> Iterator[Union[Tuple[DereferencedBlock, bytes], BaseException]]:
        try:
            while not self._reads_finished():
                logger.debug('Read queue length, outstanding reads, completion queue length: {}, {}, {}.'.format(
                    len(self._read_queue), self._outstanding_aio_reads, self._read_completion_queue.qsize()))
                self._submit_aio_reads()

                completion, t1, t2, block, data = self._read_completion_queue.get(block=(timeout is None or
                                                                                         timeout != 0),
                                                                                  timeout=timeout)
                assert self._outstanding_aio_reads > 0
                self._outstanding_aio_reads -= 1

                try:
                    completion.wait_for_complete_and_cb()
                except Exception as exception:
                    yield exception
                else:
                    read_return_value = completion.get_return_value()

                    if read_return_value < 0:
                        raise IOError('Read of block {} failed.'.format(block.idx))

                    if read_return_value != block.size:
                        raise IOError('Short read of block {}. Wanted {} bytes but got {}.'.format(
                            block.idx, block.size, read_return_value))

                    if not data:
                        # We shouldn't get here because a failed read should be caught by the "read_return_value < 0"
                        # check above. See: https://github.com/ceph/ceph/blob/880468b4bf6f0a1995de5bd98c09007a00222cbf/src/pybind/rbd/rbd.pyx#L4145.
                        raise IOError('Read of block {} failed.'.format(block.idx))

                    logger.debug('Read block {} in {:.3f}s'.format(block.idx, t2 - t1))

                    yield block, data

                self._read_completion_queue.task_done()
        except queue.Empty:
            return
        else:
            return

    def _submit_aio_writes(self):
        assert self._rbd_image is not None
        while len(self._write_queue) > 0 and self._outstanding_aio_writes < self._simultaneous_writes:
            block, data = self._write_queue.pop()
            t1 = time.time()

            def aio_callback(completion):
                t2 = time.time()
                self._write_completion_queue.put((completion, t1, t2, block))
                self._submitted_aio_writes.release()

            self._submitted_aio_writes.acquire()
            offset = block.idx * self.block_size
            # LIBRADOS_OP_FLAG_FADVISE_DONTNEED: Indicates read data will not be accessed in the near future (by anyone)
            # LIBRADOS_OP_FLAG_FADVISE_NOCACHE: Indicates read data will not be accessed again (by *this* client)
            self._rbd_image.aio_write(data, offset, aio_callback, rados.LIBRADOS_OP_FLAG_FADVISE_NOCACHE)
            self._outstanding_aio_writes += 1

    def write(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        assert self._rbd_image is not None
        self._write_queue.appendleft((block.deref(), data))
        self._submit_aio_writes()

    def write_sync(self, block: Union[DereferencedBlock, Block], data: bytes) -> None:
        assert self._rbd_image is not None
        offset = block.idx * self.block_size
        t1 = time.time()
        written = self._rbd_image.write(data, offset, rados.LIBRADOS_OP_FLAG_FADVISE_DONTNEED)
        t2 = time.time()

        logger.debug('Wrote block {} in {:.3f}s'.format(block.idx, t2 - t1))

        assert written == block.size

    def _writes_finished(self) -> bool:
        return len(self._write_queue) == 0 and self._outstanding_aio_writes == 0

    def write_get_completed(self, timeout: Optional[int] = None) -> Iterator[Union[DereferencedBlock, BaseException]]:
        try:
            while not self._writes_finished():
                logger.debug('Write queue length, outstanding writes, completion queue length: {}, {}, {}.'.format(
                    len(self._write_queue), self._outstanding_aio_writes, self._write_completion_queue.qsize()))
                self._submit_aio_writes()

                completion, t1, t2, block = self._write_completion_queue.get(block=(timeout is None or timeout != 0),
                                                                             timeout=timeout)
                assert self._outstanding_aio_writes > 0
                self._outstanding_aio_writes -= 1

                try:
                    completion.wait_for_complete_and_cb()
                except Exception as exception:
                    yield exception
                else:
                    write_return_value = completion.get_return_value()
                    if write_return_value != 0:
                        raise IOError('Write of block {} failed.'.format(block.idx))

                    logger.debug('Wrote block {} in {:.3f}s'.format(block.idx, t2 - t1))

                    yield block

                self._write_completion_queue.task_done()
        except queue.Empty:
            return
        else:
            return



import json
import logging
import re
import subprocess
from json import JSONDecodeError
from typing import Dict, List, Union

from benji.helpers.settings import benji_log_level

logger = logging.getLogger()


def setup_logging() -> None:
    # Don't raise exceptions occurring during logging
    logging.raiseExceptions = False
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(benji_log_level)


def _one_line_stderr(stderr: str):
    stderr = re.sub(r'\n(?!$)', ' | ', stderr)
    stderr = re.sub(r'\s+', ' ', stderr)
    return stderr


def subprocess_run(args: List[str],
                   input: str = None,
                   timeout: int = None,
                   decode_json: bool = False) -> Union[Dict, List, str]:
    logger.debug('Running process: {}'.format(' '.join(args)))
    try:

        result = subprocess.run(args=args,
                                input=input,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                encoding='utf-8',
                                errors='ignore',
                                timeout=timeout)
    except subprocess.TimeoutExpired as exception:
        stderr = _one_line_stderr(exception.stderr)
        raise RuntimeError(f'{args[0]} invocation failed due to timeout with output: ' + stderr) from None
    except Exception as exception:
        raise RuntimeError(f'{args[0]} invocation failed with a {type(exception).__name__} exception: {str(exception)}') from None

    if result.stderr != '':
        for line in result.stderr.splitlines():
            logger.info(line)

    if result.returncode == 0:
        logger.debug('Process finished successfully.')
        if decode_json:
            try:
                stdout_json = json.loads(result.stdout)
            except JSONDecodeError:
                raise RuntimeError(f'{args[0]} invocation was successful but did not return valid JSON. Output on stderr was: {_one_line_stderr(result.stderr)}.')

            if stdout_json is None or not isinstance(stdout_json, (dict, list)):
                raise RuntimeError(f'{args[0]} invocation was successful but did return null or empty JSON dictonary. Output on stderr was: {_one_line_stderr(result.stderr)}.')

            return stdout_json
        else:
            return result.stdout
    else:
        raise RuntimeError(f'{args[0]} invocation failed with return code {result.returncode} and output: {_one_line_stderr(result.stderr)}')
